#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.
#
#    The code in this file was modifed/derived from the elasticnodes.py
#    example with the license:
#############################################################################
##
## Copyright (C) 2010 Riverbank Computing Limited.
## Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
## All rights reserved.
##
## This file is part of the examples of PyQt.
##
## $QT_BEGIN_LICENSE:BSD$
## You may use this file under the terms of the BSD license as follows:
##
## "Redistribution and use in source and binary forms, with or without
## modification, are permitted provided that the following conditions are
## met:
##   * Redistributions of source code must retain the above copyright
##     notice, this list of conditions and the following disclaimer.
##   * Redistributions in binary form must reproduce the above copyright
##     notice, this list of conditions and the following disclaimer in
##     the documentation and/or other materials provided with the
##     distribution.
##   * Neither the name of Nokia Corporation and its Subsidiary(-ies) nor
##     the names of its contributors may be used to endorse or promote
##     products derived from this software without specific prior written
##     permission.
##
## THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
## "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
## LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
## A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
## OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
## SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
## LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
## DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
## THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
## (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
## OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE."
## $QT_END_LICENSE$
##
#############################################################################

import gc
import os
import sys
import copy
import math
import time
import random


# gpi
import gpi
from gpi import QtCore, QtGui, QtWidgets
from .associate import Bindings, isGPIAssociatedFile, isGPIAssociatedExt
from .canvasScene import CanvasScene
from .cmd import Commands
from .defines import GPI_REQUEUE_EVENT, GPI_INIT_EVENT, GPI_WIDGET_EVENT
from .defines import getKeyboardModifiers, printMouseEvent, stw
from .defines import isMacroChildNode
from .defines import GetHumanReadable_bytes, GPI_APPLOOP, GetHumanReadable_time
from .defines import isGPINetworkFile, isGPIModFile
from .edge import Edge
from .layoutWindow import LayoutMaster
from .library import Library, NodeCatalogItem
from .macroNode import MacroNode
from .network import Network
from .node import Node
from .nodeQueue import GPINodeQueue
from .port import Port, InPort
from .stateMachine import GPI_FSM, GPIState
from . import topsort

from .logger import manager

# start logger for this module
log = manager.getLogger(__name__)

class GraphWidget(QtWidgets.QGraphicsView):
    '''Provides the main canvas widget and background painting as well as the
    execution model for the canvas.'''

    changed = gpi.Signal(QtCore.QMimeData)
    _switchSig = gpi.Signal(str)
    _switchSig_info = gpi.Signal(dict)
    _curState = gpi.Signal(dict)

    def __init__(self, title, parent):
        super(GraphWidget, self).__init__()

        # a link to the main window
        self.parent = parent
        self._title = title
        self._macroModule = False

        self.hotkeys = {}
        
        # canvas info
        self._starttime = 0
        self._walltime = 0  # time between idle states

        # node animation
        self._node_anim_timeline = None
        self._node_anims = []

        self._layoutwindowList = []

        self._proc = None  # reference point for threads

        self.timerId = 0
        self.nodeEvent_timerId = self.startTimer(1000)  # update time (msec)
        self.chargeRepON = False  # start off this way

        scene = CanvasScene(self)
        scene.setItemIndexMethod(QtWidgets.QGraphicsScene.NoIndex)
        ncscale = 4  # Network Canvas Size Scale
        scene.setSceneRect(
            -200 * ncscale, -200 * ncscale, 400 * ncscale, 400 * ncscale)
        self.setScene(scene)
        self.setCacheMode(QtWidgets.QGraphicsView.CacheNone)  # required for repainting background
        self.setViewportUpdateMode(
            QtWidgets.QGraphicsView.BoundingRectViewportUpdate)
        self.setRenderHint(QtGui.QPainter.Antialiasing)
        self.setTransformationAnchor(QtWidgets.QGraphicsView.AnchorUnderMouse)
        self.setResizeAnchor(QtWidgets.QGraphicsView.AnchorViewCenter)
        self.setInteractive(True)

        self.scale(2.0, 2.0)
        self.setMinimumSize(400, 400)
        self.setDragMode(self.ScrollHandDrag)
        self._panning = False

        self.setAcceptDrops(True)
        self.setCursor(QtCore.Qt.OpenHandCursor)
        self.gridRes = 5  # pts
        self.nodeQueue = GPINodeQueue()
        self.extWidgets = dict()

        # timed painter update
        self._timer = QtCore.QTimer()
        self._timer.timeout.connect(self.viewAndSceneForcedUpdate)
        # self._timer.start(1000) # 10msec update

        # TODO: this probably should go to the MainCanvas
        self._library = Library(self)
        #self._library.scanGPIModulesIn_SysPath(recursion_depth=2)
        #self._library.generateLibMenus()

        self._event_pos = QtCore.QPoint(0, 0)

        self._network = Network(self)

        self._pause_quiet = False

        self.initStateMachine()

    def rescanLibrary(self):
        self._library.scanForNewNodes()

    def getLibrary(self):
        return self._library

    def getEventPos(self):
        return self._event_pos

    def getEventPos_randomDev(self, rad=None):
        if rad:
            radius = rad
        else:
            radius = 10.0  # pts
        x = self._event_pos.x() + random.random() * radius
        y = self._event_pos.y() + random.random() * radius
        pos = QtCore.QPoint(x, y)
        pos = self.mapToScene(pos)
        return pos

    def title(self):
        return self._title

    def initStateMachine(self):  # GRAPH
        # Set up intial state graph.
        self._machine = GPI_FSM('GRAPH')
        self._switchSig.connect(self._machine.next)
        self._switchSig_info.connect(self._machine.next)

        # node states
        self._undefinedStateSig = {'title':self.title(), 'msg':'Undefined State (how did you get here?)'}
        self._initState = GPIState('init', self.initRun, self._machine, efunc=self.initWalltime)
        self._idleState = GPIState('idle', self.idleRun, self._machine, efunc=self.initWalltime)
        self._idleStateSig = {'title':self.title(), 'msg':'Idle'}
        self._checkEventsState = GPIState('checkEvents', self.checkEventsRun, self._machine)
        self._checkEventsStateSig = {'title':self.title(), 'msg':'Checking Events'}
        #self._deleteNodeState = GPIState('deleteNode', self.deleteNodeRun,
        #        self._machine)
        #self._addNodeState = GPIState('addNode', self.addNodeRun,
        #        self._machine)
        self._processingState = GPIState('processing', self.processingRun,
                self._machine)
        self._processingStateSig = {'title':self.title(), 'msg':'Processing'}
        self._pausedState = GPIState('paused', self.pausedRun, self._machine, efunc=self.pausedLeave)
        self._pausedStateSig = {'title':self.title(), 'msg':'Paused'}

        # make state graph
        # init
        self._initState.addTransition('init_check', self._checkEventsState)
        self._initState.addTransition('init_finished', self._idleState)
        self._initState.addTransition('pause', self._pausedState)
        #self._initState.exited.connect(self.initWalltime)

        # idle
        self._idleState.addTransition('check', self._checkEventsState)
        #self._idleState.addTransition('delete', self._deleteNodeState)
        #self._idleState.addTransition('deleteAll', self._deleteNodeState)
        #self._idleState.addTransition('load', self._addNodeState)
        self._idleState.addTransition('pause', self._pausedState)
        #self._idleState.exited.connect(self.initWalltime)

        # checkEvents
        self._checkEventsState.addTransition('process', self._processingState)
        self._checkEventsState.addTransition('requeue', self._checkEventsState)
        #self._checkEventsState.addTransition('delete', self._deleteNodeState)
        #self._checkEventsState.addTransition('deleteAll',
        #                                     self._deleteNodeState)
        self._checkEventsState.addTransition('ignore', self._idleState)
        #self._checkEventsState.addTransition('load', self._addNodeState)
        self._checkEventsState.addTransition('pause', self._pausedState)

        # deleteNode
        #self._deleteNodeState.addTransition('check', self._checkEventsState)
        #self._deleteNodeState.addTransition('process', self._processingState)

        # addNode
        #self._addNodeState.addTransition('check', self._checkEventsState)
        #self._addNodeState.addTransition('afterload', self._processingState)

        # processing
        #self._processingState.addTransition('delete', self._deleteNodeState)
        #self._processingState.addTransition('deleteAll', self._deleteNodeState)
        self._processingState.addTransition('pause', self._pausedState)
        self._processingState.addTransition('check', self._checkEventsState)
        self._processingState.addTransition('next', self._processingState)
        #self._processingState.addTransition('load', self._addNodeState)

        # pause
        self._pausedState.addTransition('unpause', self._checkEventsState)

        #self._machine.start(self._idleState)
        self._machine.start(self._initState)

    def walltime(self):
        return self._walltime

    def clearWalltime(self):
        if 'walltime' in self._idleStateSig:
            self._idleStateSig.pop('walltime')

    def initWalltime(self, sig):
        # sig is a dummy so that it can be an onExit state transition
        self._starttime = time.time()

    def calcWalltime(self):
        self._walltime = time.time() - self._starttime

    def walltime_disp(self):
        return GetHumanReadable_time(self.walltime(), precision=1)

    def initRun(self, sig):
        # run any initialization stuff here
        # since the 'check state' can't run yet, the canvas is virtually paused.

        if Commands.pendingCount():

            # load networks
            if Commands.netCount():
                for path in Commands.nets():

                    pos = self.getEventPos_randomDev(rad=50)
                    pos = QtCore.QPoint(pos.x(), pos.y())

                    s = {'sig': 'load', 'subsig': 'net', 'path':
                            path, 'pos': pos}
                    self.addNodeRun(s)

            # load nodes
            if Commands.modCount():
                for path in Commands.mods():

                    pos = self.getEventPos_randomDev(rad=50)
                    pos = QtCore.QPoint(pos.x(), pos.y())

                    s = {'sig': 'load', 'subsig': 'mod',
                            'path': path, 'pos': pos, 'from': 'cmd.Commands'}
                    self.addNodeRun(s)

            # load associated files
            if Commands.fileCount():
                for path in Commands.files():

                    pos = self.getEventPos_randomDev(rad=50)
                    pos = QtCore.QPoint(pos.x(), pos.y())

                    bpath, file_ext = os.path.splitext(path)
                    s = {'sig': 'load', 'subsig': file_ext, 'path': path, 'pos': pos}
                    self.addNodeRun(s)

            # NOTE: macro-nodes that need to close, re-select themselves
            #   -so this call doesn't work on them
            self.scene().unselectAllItems()

            # once all networks are loaded, process node arguments

            # String-Node Args
            if Commands.stringNodeArgCount():
                for lab in Commands.stringNodeLabels():
                    node = self.findNodeByNameAndLabel('String', lab)
                    if node:
                        # get the string arg
                        arg = Commands.stringNodeArg(lab)

                        # set 'string' widget value
                        node._nodeIF.modifyWidget_direct('string', val=arg)
                        node.setEventStatus({GPI_WIDGET_EVENT: 'string'})
                    else:
                        log.warn('String node label: \''+str(lab)+'\' not found, skipping.')

            self._switchSig.emit('init_check')

        else:
            self._switchSig.emit('init_finished')

    def totalPortMem(self):
        bytes_held = 0
        for node in self.getAllNodes():
            bytes_held += node.portMem()
        return bytes_held

    def totalPortMem_disp(self, bytes_held):
        return 'Total Port MEM: '+GetHumanReadable_bytes(bytes_held)

    # Function executed upon state change:
    def idleRun(self, sig):

        # get walltime and put it here
        self.calcWalltime()
        if self.walltime() > 0:
            self._idleStateSig['walltime'] = self.walltime_disp()
        else:
            self.clearWalltime()

        self._curState.emit(self._idleStateSig)
        self.printCurState()
        self.viewAndSceneForcedUpdate()

        # idle is a good time to force collection
        log.debug('pausedRun(): garbage collect')
        gc.collect()

        # if GPI was started without GUI, then assume the network has finished and exit
        if Commands.noGUI() or Commands.scriptMode():
            self.deleteAllNodeMMAPs()
            log.dialog('Canvas Wall Time: '+str(self.walltime_disp()) + ', exiting.')
            sys.exit(0)

    def pausedRun(self, sig):
        self._curState.emit(self._pausedStateSig)  # update statusbar
        self.printCurState()

        # don't draw yellow bkgnd
        if 'subsig' in sig:
            self._pause_quiet = True

        self.viewAndSceneForcedUpdate()

        # pause is a good time to force collection
        log.debug('pausedRun(): garbage collect')
        gc.collect()

        # if GPI was started without GUI, then assume the network has finished and exit
        if Commands.noGUI() or Commands.scriptMode():
            self.deleteAllNodeMMAPs()
            log.dialog('The canvas fell into a paused state, exiting.')
            sys.exit(1)

    def pausedLeave(self, sig):
        # always reset quiet flag
        self._pause_quiet = False

    def checkEventsRun(self, sig):
        self._curState.emit(self._checkEventsStateSig)
        self.printCurState()

        # Currently Running nodes
        if self.aNodeIsProcessing():
            self._switchSig.emit('process')
            return

        # EVENTS
        # check for event status BEFORE triggering highest compute
        for node in self.getAllNodes():
            if node.isReady():
                # Re/-initialize queue and start processing.
                # This was called because 'a' node has an event status.
                self.nodeQueue.setQueue(self.getLinearNodeHierarchy())
                self._switchSig.emit('process')
                return

        # REQUEUE EVENTS
        # if queue is done then check for re-queue nodes
        if self.nodeQueue.isEmpty():
            log.debug("checkEventsRun(): check for requeue nodes.")
            nodes = self.getAllNodes()
            cnt = 0
            for node in nodes:
                if node._nodeIF:  # protect against deleted object
                    if node._nodeIF.reQueueIsSet() and \
                            not node.inDisabledState():
                        node.setEventStatus({GPI_REQUEUE_EVENT: None})
                        cnt += 1
            if cnt:  # if any nodes got reset then start loop
                self._switchSig.emit('requeue')
                return

        # NO EVENTS
        # else: no events or requeue events were found
        self._switchSig.emit('ignore')

    def newNode_byClosestMatch(self, name, wdg_port_names, pos, mapit=False):
        # Search the library for all nodes with the same name, then do a
        # sub-search based on a list of wdg names.
        item = self._library.findNode_byClosestMatch(name, wdg_port_names)
        if item:
            item.reload()
            log.debug('\tfound')
            return self.newNode_byNodeCatalogItem(item, pos, mapit)
        else:
            log.debug('\tfailed to find node')

    def newNode_byKey(self, key, pos, mapit=False):
        # search the library for a node with given name
        item = self._library.findNode_byKey(key)
        if item:
            item.reload()
            log.debug('\tfound')
            return self.newNode_byNodeCatalogItem(item, pos, mapit)
        else:
            log.debug('\tfailed to find node')

    def newNode_byName(self, name, pos, mapit=False):
        # search the library for a node with given name
        item = self._library.findNode_byName(name)
        if item:
            item.reload()
            log.debug('\tfound')
            return self.newNode_byNodeCatalogItem(item, pos, mapit)
        else:
            log.debug('\tfailed to find node')

    def newNode_byPath(self, path, pos, mapit=False):
        # just try to make a node item, if it loaded then its valid.
        item = NodeCatalogItem(path)
        item.load()
        if item.valid():
            log.debug('\tsuccess')
            return self.newNode_byNodeCatalogItem(item, pos, mapit)
        else:
            log.debug('\titem cannot be loaded')

    def newNode_byNodeCatalogItem(self, item, pos, mapit=False):
        '''Add a new node to the canvas from a NodeCatalogItem description.
        Return a handle to the new canvas item.
            pos: QtCore.QPoint()
        '''
        if item is None:
            return None

        # Update user modifications (if any).
        item.reload()

        # If the user has made changes that cause the node to be non-loadable
        # then return None
        if not item.valid():
            return None

        newnode = Node(self, nodeCatItem=item)

        # force all execType(s) to be GPI_APPLOOP
        if False:  # letting them all be processes seems to be the safest for now
        #if Commands.noGUI():
            # Thread seems safer, APPLOOP was causing recursion errors.
            # Probably due to signals piling up.

            # GPI_APPLOOP & iter_test.net causes: recursion error
            et = lambda :GPI_APPLOOP

            # GPI_THREAD & beta_spiral.net causes: 64119 Bus error: 10
            #et = lambda :GPI_THREAD

            newnode.execType = et
            newnode._nodeIF.execType = et

        newnode.refreshName()
        self.scene().addItem(newnode)
        if mapit:
            mpos = self.mapToScene(pos)
        else:
            mpos = pos
        newnode.setPos(mpos.x(), mpos.y())
        return newnode

    # TODO: since addNodeRun was removed from the state-machine it now needs
    # a real function interface instead of passing a dict to parameterize
    def addNodeRun(self, sig):  # state: 'addNode', 'Run' method
        self.printCurState()

        if type(sig['subsig']) == NodeCatalogItem:
            log.debug('addNode by item')
            item = sig['subsig']

            # get the position of menu invocation
            radius = 10.0  # pts
            if "pos" not in sig.keys():
                x = self._event_pos.x() + random.random() * radius
                y = self._event_pos.y() + random.random() * radius
                pos = QtCore.QPoint(x, y)
            else:
                pos = sig['pos']

            # instantiate node on canvas
            if "mapit" not in sig.keys():
                mapit = True
            else:
                mapit = sig['mapit']
            node = self.newNode_byNodeCatalogItem(item, pos, mapit)
            if node:
                self.scene().makeOnlyTheseNodesSelected([node])
                node.setEventStatus({GPI_INIT_EVENT: None})
                self.ensureVisible(node)

        elif sig['subsig'] == 'mod':
            log.debug('addNode by path')
            path = sig['path']
            pos = sig['pos']

            # instantiate node on canvas
            node = self.newNode_byPath(path, pos, mapit=True)
            if node:
                self.scene().makeOnlyTheseNodesSelected([node])
                node.setEventStatus({GPI_INIT_EVENT: None})
                self.ensureVisible(node)

        # 3-4 pieces of info for file associations
        #   node-name (and possibly key), file ext, string widget to push to
        elif isGPIAssociatedExt(sig['subsig']):

            # get binding for this file extension
            # all extensions should be case-insensitive
            item = Bindings.get(sig['subsig'].lower())

            # assume the item is holding a full key
            node = self.newNode_byKey(item.node, sig['pos'], mapit=True)
            if node is None:
                node = self.newNode_byName(item.node, sig['pos'], mapit=True)
            if node is None:
                log.error('\''+str(item.node)+'\' could not be located for \''+str(item.ext)+'\'')
            else:
                node._nodeIF.modifyWidget_direct(item.wdg, val=sig['path'])
                self.scene().unselectAllItems()
                node.setSelected(True)
                node.setEventStatus({GPI_WIDGET_EVENT: item.wdg})
                self.ensureVisible(node)

        elif sig['subsig'] == 'net':

            if 'pos' in sig:
                net = self._network.loadNetworkFromFile(sig['path'])
                if net:
                    self.deserializeCanvas(net, sig['pos'])
            else:
                net = self._network.loadNetworkFromFile(sig['path'])
                if net:
                    self.deserializeCanvas(net, self.getEventPos_randomDev())

        elif sig['subsig'] == 'dialog':
            if 'pos' in sig:
                net = self._network.loadNetworkFromFileDialog()
                if net:
                    self.deserializeCanvas(net, sig['pos'])
            else:
                net = self._network.loadNetworkFromFileDialog()
                if net:
                    self.deserializeCanvas(net, self.getEventPos_randomDev())

        elif sig['subsig'] == 'paste':
            if self.parent._copybuffer:
                self.deserializeGraphData(self.parent._copybuffer, pos=sig['pos'])

        elif sig['subsig'] == 'keypaste':
            if self.parent._copybuffer:
                self.deserializeGraphData(self.parent._copybuffer, offset=True, randoffset=True)

        elif sig['subsig'] == 'reload':
            if self.parent._copybuffer:
                self.deserializeGraphData(self.parent._copybuffer, reloadnode=True)

        if self.inIdleState():# or self.inCheckEventsState():
            self._switchSig.emit('check')

        return node

    def deleteNodeRun(self, sig):
        self.printCurState()
        if sig == 'delete':  # delete selected
            self.deleteSelectedNodes()
        elif sig == 'deleteAll':  # clear all
            self.deleteAllNodes()

        # back to processing or check for new events
        #if self.nodeQueue.isEmpty():
        #    self._switchSig.emit('check')
        #else:
        #    self._switchSig.emit('process')

        if self.inIdleState():
            self._switchSig.emit('check')

    def aNodeIsProcessing(self):
        for node in self.getAllNodes():
            if node.isProcessingEvent():
                return True
        return False

    def processingRun(self, sig):
        self._curState.emit(self._processingStateSig)
        self.printCurState()

        if not self.aNodeIsProcessing():
            queueState = self.nodeQueue.startNextNode()
            if queueState == 'paused':
                self._switchSig.emit('paused')
            elif queueState == 'finished':
                self._switchSig.emit('check')

        self.viewAndSceneForcedUpdate()

    # State Checking:
    def getCurState(self):
        return self._machine.curState

    def getCurStateName(self):
        '''return state names in a list of strings'''
        return self._machine.curStateName

    def getCurStateSig(self):
        if self.inIdleState():
            return self._idleStateSig
        elif self.inPausedState():
            return self._pausedStateSig
        elif self.inCheckEventsState():
            return self._checkEventsStateSig
        elif self.inProcessingState():
            return self._processingStateSig
        else:
            return self._undefinedStateSig

    def printCurState(self):
        log.debug("GRAPH State(s): "+self.getCurStateName())

    def inIdleState(self):  # GRAPH
        return self._idleState is self.getCurState()

    def inPausedState(self):  # GRAPH
        return self._pausedState is self.getCurState()

    def inCheckEventsState(self):
        return self._checkEventsState is self.getCurState()

    def inProcessingState(self):
        return self._processingState is self.getCurState()

    def printNodeState(self):
        allItems = self.getAllNodes()
        for node in allItems:
            print("________________________")
            node.printCurState()
            print(("node: " + str(node.name)))
            print(("inDisabledState: " + str(node.inDisabledState())))
            print(("hasEventPending: " + str(node.hasEventPending())))
            print(("_nodeIF.reQueueIsSet: " + str(node._nodeIF.reQueueIsSet())))
            print("________________________")

    def setPauseState(self, val):
        old_val = self.nodeQueue.isPaused()
        self.nodeQueue.setPause(val)
        if val != old_val:  # state changed
            if not val:  # unpaused
                # after pause drop old event queue
                self.nodeQueue.resetQueue()
                self._switchSig.emit('check')

    def isPaused(self):
        return self.nodeQueue.isPaused()

    def scrollContentsBy(self, x, y):
        super(GraphWidget, self).scrollContentsBy(x, y)
        y = self.geometry().height() / 2
        x = self.geometry().width() / 2
        self._event_pos = QtCore.QPoint(x, y)

    def newLayoutWindowFromSettings(self, s, nodeList):

        # config has to be set at construction b/c layouts cannot yet be
        # deleted.
        layoutwindow = LayoutMaster(self, config=s['config'])
        layoutwindow.loadSettings(s, nodeList)
        layoutwindow.setWindowTitle(self._title+".Layout Window "+str(len(self._layoutwindowList)+1))

        #scrollArea = QtWidgets.QScrollArea()
        #scrollArea.setWidget(layoutwindow)
        #scrollArea.setWidgetResizable(True)
        #scrollArea.setGeometry(50, 50, 300, 1000)
        #self._layoutwindowList.append(scrollArea)

        self._layoutwindowList.append(layoutwindow)
        layoutwindow.setGeometry(50, 50, 400, 300)

        #scrollArea.show()
        #scrollArea.raise_()

        layoutwindow.show()
        layoutwindow.raise_()

    def newLayoutWindow(self, config):

        layoutwindow = LayoutMaster(self, config=config)
        layoutwindow.setWindowTitle(self._title+".Layout Window "+str(len(self._layoutwindowList)+1))

        #scrollArea = QtWidgets.QScrollArea()
        #scrollArea.setWidget(layoutwindow)
        #scrollArea.setWidgetResizable(True)
        #scrollArea.setGeometry(50, 50, 300, 1000)
        #self._layoutwindowList.append(scrollArea)

        self._layoutwindowList.append(layoutwindow)
        layoutwindow.setGeometry(50, 50, 400, 300)

        #scrollArea.show()
        #scrollArea.raise_()

        layoutwindow.show()
        layoutwindow.raise_()

    def serializeLayoutWindows(self):
        '''Save each layout dict.
        '''
        s = []
        for layout in self._layoutwindowList:
            # layout is None for past closed windows.
            # this keeps numbering correct
            if layout is not None:
                s.append(layout.getSettings())
        return s


    def dragEnterEvent(self, event):
        if event.mimeData().hasFormat('text/uri-list'):
            event.acceptProposedAction()
        self.changed.emit(event.mimeData())

    def dragMoveEvent(self, event):
        if event.mimeData().hasFormat('text/uri-list'):
            event.acceptProposedAction()

    def dropEvent(self, event):

        if event.mimeData().hasFormat('application/gpi-widget'):
            mime = event.mimeData()
            itemData = mime.data('application/gpi-widget')
            dataStream = QtCore.QDataStream(
                itemData, QtCore.QIODevice.ReadOnly)

            text = QtCore.QByteArray()
            offset = QtCore.QPoint()
            dataStream >> text >> offset

            log.debug("canvasGraph(): Mime data:")
            log.debug(str(text))
            log.debug(str(offset))
            return

        elif event.mimeData().hasFormat('text/uri-list'):
            mimeData = event.mimeData()

            log.debug(str(mimeData))
            paths = [str(x.path()) for x in mimeData.urls()]
            log.debug(paths)

            # if multiple drops, then add random offsets to pos
            if len(paths) == 1:
                poses = [event.pos()]
            else:
                poses = []
                for path in paths:
                    m = 50
                    rand = QtCore.QPoint(random.random()*m, random.random()*m)
                    poses.append(event.pos() + rand)

            # process each dropped path
            for path, pos in zip(paths, poses):

                log.debug('Dropped uri: '+str(path))

                # node definitions
                if isGPIModFile(path):

                    # add the node to the library (and menu) if possible
                    item = NodeCatalogItem(path)
                    ret = self._library.addNode(item)
                    if ret > 0:
                        log.dialog('Added dropped node to the library.')
                        self._library.regenerateLibMenus()
                    elif ret == 0:
                        log.dialog('Dropped Node is already in the library.')
                    else:
                        log.error('Dropped Node is Invalid.')
                        return

                    # add the node to the canvas
                    s = {'sig': 'load', 'subsig': 'mod',
                            'path': path, 'pos': pos, 'from': 'Dropped uri'}
                    self.addNodeRun(s)

                # file associations
                elif isGPIAssociatedFile(path):
                    bpath, file_ext = os.path.splitext(path)
                    s = {'sig': 'load', 'subsig': file_ext, 'path': path, 'pos': pos}
                    self.addNodeRun(s)

                # network files
                elif isGPINetworkFile(path):
                    s = {'sig': 'load', 'subsig': 'net', 'path':
                            path, 'pos': self.mapToScene(pos)}
                    self.addNodeRun(s)

                # not a recognized file
                else:
                    log.warn("dropEvent(): Filetype not recognized by GPI.")
                    return

            # shows rejected animation if not called
            event.acceptProposedAction()

    def dragLeaveEvent(self, event):
        event.accept()

    def itemMoved(self):

        # only start this timer if the menu option for CR is set
        if self.chargeRepON:
            if not self.timerId:
                self.timerId = self.startTimer(30)  # update time (msec)

                # add timeout
                # give object a few seconds to unwrap
                # self.stimer = QtCore.QTimer()
                # self.stimer.singleShot(10000, self.send_killTimer)

    def deleteNode(self, node):

        if isinstance(node, Node):
            #if (node.execType() is GPI_THREAD) and node.isProcessingEvent():
            #    print "Thread is in progress, cancel delete ("+node.name+")"
            #    return
            if isMacroChildNode(node):
                node.macroParent().readyForDeletion()
                for n in node.getSiblingNodes():
                    self.nodeQueue.removeNode(n)
                    n.readyForDeletion()
                    if n.scene():
                        self.scene().removeItem(n)
            elif node:
                node.readyForDeletion()
                if node.scene():
                    self.scene().removeItem(node)

        # keep random objects from being copied to other processes
        log.debug('deleteNode(): garbage collect')
        gc.collect()

        # try to check check for changes after a deletion
        if self.inProcessingState():
            self._switchSig.emit('check')

    def deleteSelectedNodes(self):
        '''For a list of nodes, its safer to disable all of them and remove
        them from the queue directly
        '''
        selnodes = self.getSelectedNodes()
        for node in selnodes:
            node.setDeleteFlag(True)
            node.setDisabledState(True)
            self.nodeQueue.removeNode(node)
        for node in selnodes:
            self.deleteNode(node)

    def deleteAllNodes(self):
        '''For a list of nodes, its safer to disable all of them and remove
        them from the queue directly
        '''
        selnodes = self.getAllNodes()
        for node in selnodes:
            node.setDeleteFlag(True)
            node.setDisabledState(True)
            self.nodeQueue.removeNode(node)
        for node in selnodes:
            self.deleteNode(node)

    def deleteAllNodeMMAPs(self):
        '''For a list of nodes, its safer to disable all of them and remove
        them from the queue directly
        '''
        selnodes = self.getAllNodes()
        for node in selnodes:
            node.removeMMAPs()

    def getAllMacroNodes(self):
        '''Find all nodes that belong to macro-framework, then store them in a
        dictionary based on macro-object-id.
        '''
        macros = {}
        mnodes = []
        for node in self.getAllNodes():
            if isMacroChildNode(node):
                macros[str(node.macroParent().getID())] = node.getSiblingNodes()
                mnodes.append(node.macroParent())

        # take encapsulated nodes if the macro is collapsed
        enodes = []
        for node in mnodes:
            if node.isCollapsed():
                enodes += node.getEncapsulatedNodes()
        enodes = list(set(enodes))

        return macros, enodes

    def getSelectedMacroNodes(self):
        '''Find all nodes that belong to macro-framework, if ANY of the child
        nodes are selected then the node is selected:
            -if its expanded, then only IT is selected (and any other selected
            nodes)
            -if its collapsed, then it AND all encapsulated nodes are selected.
        '''
        macros = {}
        mnodes = []
        for node in self.getSelectedNodes():
            if isMacroChildNode(node):
                macros[str(node.macroParent().getID())] = node.getSiblingNodes()
                mnodes.append(node.macroParent())

        # take encapsulated nodes if the macro is collapsed
        enodes = []
        for node in mnodes:
            if node.isCollapsed():
                enodes += node.getEncapsulatedNodes()
        enodes = list(set(enodes))

        return macros, enodes

    def getAllNodes(self):
        allitems = list(self.scene().items())[:]  # copy in case of user interrupt
        nodes = [item for item in allitems if isinstance(item, Node)]
        return(nodes)

    def getAllMacros(self):
        '''Get the MacroNode object class handle.
        '''
        allitems = list(self.scene().items())[:]  # copy in case of user interrupt
        nodes = [item for item in allitems if isinstance(item, MacroNode)]
        return(nodes)

    def findNodeByNameAndLabel(self, name, lab):
        # return the first occurrence of a Node with the given name and label
        for node in self.getAllNodes():
            if node.getNodeLabel() == lab:  # most exclusive
                if node.getNameFromItem() == name:
                    return node

    def getAllPorts(self):
        allitems = list(self.scene().items())[:]  # copy in case of user interrupt
        ports = [item for item in allitems if isinstance(item, Port)]
        return(ports)

    def getSelectedNodes(self):
        sceneItems = list(self.scene().items())[:]  # copy in case of user interrupt
        sceneItems = [
            i for i in sceneItems if i.isSelected() and isinstance(i, Node)]
        return sceneItems

    def getEmptyConnectionNodes(self, nodes):
        empty = []
        for node in nodes:
            connections = node.getOutputConnections()
            if (not any(connections)): empty.append(node)
        return empty

    def findWidgetByID(self, nodeList, wdgid):
        '''traverses all node's parmLists for the given id.
        '''
        wdgid = int(wdgid)
        for node in nodeList:
            for parm in node._nodeIF.parmList:
                if parm.get_id() == wdgid:
                    return parm

    # kill timer for items moving under charge repulsion
    def send_killTimer(self):
        self.killTimer(self.timerId)
        self.timerId = 0
        self.chargeRepON = False

    def getLinearNodeHierarchy(self):
        return self.getLinearNodeHierarchy_fromList(self.getAllNodes())

    def getLinearNodeHierarchy_fromList(self, nodeList):
        return sorted(nodeList, key=lambda y: y.getHierarchalLevel())

    def viewAndSceneForcedUpdate(self):
        '''All calls to this updater are to patch over a bug that
        presents when a network has iterated too many times.
        -The root problem needs to be found (perhaps pyqt vs. qt ownership).

        -This is also required to immediately render the pause event.
        '''
        log.debug("viewAndSceneForcedUpdate called")
        ##self.updateMicroFocus()
        ##self.updateGeometry()
        ##self.repaint()

        # don't bother updating if there is no gui
        if Commands.noGUI():
            return

        self.update()
        self.scene().update()

        ##QtWidgets.QApplication.processEvents() # allow gui to update

    def calcNodeHierarchy(self):
        # tells each node which level it is
        # and returns a list based on that level

        nodeList = self.getAllNodes()

        # concatenate all connections (even if list is redundant)
        c = []
        for node in nodeList:
            if len(node.getNonCyclicConnectionTuples()):
                c += node.getNonCyclicConnectionTuples()
            #if len(node.getConnectionTuples()):
            #    c += node.getConnectionTuples()
            else:
                # island nodes have top priority
                node.resetHierarchalLevel()
                node.refreshName()

        sortedNodes = topsort.topsort(c)

        # signal that the connection is cyclic
        if sortedNodes is None:
            return None

        # set node hierarchy
        # -each node knows its current level
        cnt = 0
        for node in sortedNodes:
            node.setHierarchalLevel(cnt)
            cnt += 1

        return sortedNodes

    def roundPosToGrid(self, pos):
        x = int(pos[0] / self.gridRes) * self.gridRes
        y = int(pos[1] / self.gridRes) * self.gridRes
        return(x, y)

    def keyPressEvent(self, event):
        key = event.key()
        modifiers = getKeyboardModifiers()

        # copy/paste/delete
        if key == QtCore.Qt.Key_C and modifiers == QtCore.Qt.ControlModifier:
            self.copyNodesToBuffer()
        elif key == QtCore.Qt.Key_V and modifiers == QtCore.Qt.ControlModifier:
            # try to paste fairly close to where the nodes were copied
            if self.parent._copybuffer:
                s = {'sig': 'load', 'subsig': 'keypaste'}
                self.addNodeRun(s)
            else:
                log.warn("Nothing in buffer to paste.")
        elif key == QtCore.Qt.Key_Delete or key == QtCore.Qt.Key_Backspace:
            #self._switchSig.emit('delete')  # change state
            self.deleteNodeRun('delete')

        # load/save network
        elif key == QtCore.Qt.Key_L and modifiers == QtCore.Qt.ControlModifier:
            s = {'sig': 'load', 'subsig': 'dialog'}
            self.addNodeRun(s)
            #self._switchSig_info.emit(s)

        elif key == QtCore.Qt.Key_S and modifiers == QtCore.Qt.ControlModifier:
            self._network.saveNetworkFromFileDialog(self.serializeCanvas())

        # move nodes across canvas
        elif key == QtCore.Qt.Key_Up:
            for node in self.getSelectedNodes():
                pos = node.getPos()
                x, y = self.roundPosToGrid(pos)
                node.moveBy(0, y - pos[1] - 5)
        elif key == QtCore.Qt.Key_Down:
            for node in self.getSelectedNodes():
                pos = node.getPos()
                x, y = self.roundPosToGrid(pos)
                node.moveBy(0, y - pos[1] + 5)
        elif key == QtCore.Qt.Key_Left:
            for node in self.getSelectedNodes():
                pos = node.getPos()
                x, y = self.roundPosToGrid(pos)
                node.moveBy(x - pos[0] - 5, 0)
        elif key == QtCore.Qt.Key_Right:
            for node in self.getSelectedNodes():
                pos = node.getPos()
                x, y = self.roundPosToGrid(pos)
                node.moveBy(x - pos[0] + 5, 0)

        # tab nodes (in execution order)
        elif key == QtCore.Qt.Key_Tab:
            snodes = self.getSelectedNodes()
            if len(snodes):  # just skip if no nodes are selected
                snode = snodes[0]
                nodes = self.getLinearNodeHierarchy()
                for i in range(len(nodes)):
                    if nodes[i] == snode:
                        if i < len(nodes) - 1:
                            self.scene().makeOnlyTheseNodesSelected(
                                [nodes[i + 1]])
                        else:
                            self.scene().makeOnlyTheseNodesSelected([nodes[0]])
                        self.scene().update()
                        return
            else:  # no nodes selected so pick one
                nodes = self.getLinearNodeHierarchy()
                self.scene().makeOnlyTheseNodesSelected([nodes[0]])
                self.scene().update()

        # raise node menu(s)
        #elif key == QtCore.Qt.Key_Space:
        #    nodes = self.getSelectedNodes()
        #    if len(nodes):
        #        for node in nodes:
        #            node.menu()

        elif key == QtCore.Qt.Key_Plus:
            self.scaleView(1.2)
        elif key == QtCore.Qt.Key_Minus:
            self.scaleView(1 / 1.2)
        elif key == QtCore.Qt.Key_Enter:
            pass
       
        # mix up nodes
        elif key == QtCore.Qt.Key_M and modifiers == QtCore.Qt.ControlModifier:
            for item in list(self.scene().items()):
                if isinstance(item, Node):
                    item.setPos(-150 + QtCore.qrand() %
                                300, -150 + QtCore.qrand() % 300)

        # organize nodes
        elif key == QtCore.Qt.Key_O and modifiers == QtCore.Qt.ControlModifier:
            self.organizeSelectedNodes()

        # pause
        elif key == QtCore.Qt.Key_P and modifiers == QtCore.Qt.ControlModifier:
            self.pauseToggle()

        # select all nodes
        elif key == QtCore.Qt.Key_A and modifiers == QtCore.Qt.ControlModifier:
            self.scene().makeOnlyTheseNodesSelected(self.getAllNodes())

        # pause -for stationary leftys
        elif key == QtCore.Qt.Key_Space:
            self.pauseToggle()

        # charge repulsion toggle
        elif key == QtCore.Qt.Key_R and int(event.modifiers()) == (QtCore.Qt.ControlModifier + QtCore.Qt.ShiftModifier):
            if self.chargeRepON is True:
                self.chargeRepON = False
            else:
                self.chargeRepON = True
                self.itemMoved()
            log.dialog("toggle chargeRepON:" + str(self.chargeRepON))

        # reload node
        elif key == QtCore.Qt.Key_R and modifiers == QtCore.Qt.ControlModifier:
            self.reload_node()

        # resize canvas window for podcast
        elif key == QtCore.Qt.Key_W and modifiers == QtCore.Qt.ControlModifier:
            log.dialog("resize window")
            self.parent.resize(1024, 768)

        # Test Key
        elif key == QtCore.Qt.Key_T:
            pass
            # log.dialog("Test Key Pressed")
            #print self.getAllPorts()
            #print self.getAllMacroNodes()
            #print self.serializeGraphData()
            # print((self.getAllNodes()))
            # print((self.getAllMacroNodes()))

        # close all node windows
        elif key == QtCore.Qt.Key_X and modifiers == QtCore.Qt.ControlModifier:
            self.closeAllNodeMenus()

        else:
            super(GraphWidget, self).keyPressEvent(event)

    def addNodeByName(self, name, pos=QtCore.QPoint(50, 35)):
        item = self._library.findNode_byName(name)
        s = {'subsig': item, 'pos': pos, 'mapit': False}
        node = self.addNodeRun(s)
        return node

    def addHotkey(self, key, node):
        hotkey = QtWidgets.QShortcut(QtGui.QKeySequence(key), self)
        hotkey.activated.connect(lambda: self.shortcut(node))
        self.hotkeys[key] = hotkey
        return hotkey


    def addShortcuts(self, shortcuts):
        for shortcut in shortcuts:
            shortcut = shortcut.split(":")
            if len(shortcut) == 2:
                self.addHotkey(shortcut[0], shortcut[1])


    def updateShortcuts(self, shortcuts):
        for key in self.hotkeys.keys():
            hotkey = self.hotkeys[key]
            hotkey.setParent(None)
        self.hotkeys = {}
        self.addShortcuts(shortcuts)
        
    def shortcut(self, name):
        # get selected nodes
        selected_nodes = self.getSelectedNodes()

        # add node and get its input ports
        node = self.addNodeByName(name, self.mousePos)
        inports = node.inportList

        # check for viable outports of the selected nodes
        viable_outports = [[] for _ in range(len(inports))]
        for s_node in selected_nodes:
            outports = s_node.outportList
            for i, inport in enumerate(inports):
                matching_ports = inport.findMatchingOutPorts(outports)
                viable_outports[i].append(matching_ports)
        
        # connect the viable outputs to the node inputs
        connected = []
        for i, ports in enumerate(viable_outports):
            inport = inports[i]

            # get ports in order, use a port from each node first
            max_l = [len(x) for x in ports]
            if len(max_l):
                max_length = max(max_l)
                temp = []
                for i in range(max_length):
                    for l in ports:
                        if i < len(l): temp.append(l[i])
                ports = [temp]

            # connect ports
            for outports in ports:
                outports = list(filter(lambda port: port not in connected, outports))
                if len(outports) and inport not in connected:
                    outport = outports[0]
                    if outport in connected: continue # skip if the outport is already connected
                    newEdge = Edge(outport, inport)
                    self.scene().addItem(newEdge)
                    connected.append(outport)
                
                    nodeHierarchy = inport.getNode().graph.calcNodeHierarchy()
                    if nodeHierarchy is None:
                        self.scene().removeItem(newEdge)
                        newEdge.detachSelf()
                        # del newEdge
                        log.warn("CanvasScene: cyclic, connection dropped")
                    else:
                        # CONNECTION ADDED
                        # Since node hierarchy is recalculated, also
                        # take the time to flag nodes for processing
                        # 1) check for matching spec type
                        if not (inport.checkUpstreamPortType()):
                            self.scene().removeItem(newEdge)
                            newEdge.detachSelf(update=False)
                            # del newEdge
                            log.warn("CanvasScene: data type mismatch, connection dropped")
                        else:
                            # 2) set the downstream node's pending_event
                            GPI_PORT_EVENT = '_PORT_EVENT_'
                            inport.getNode().setEventStatus({GPI_PORT_EVENT: inport.portTitle})

                            # trigger a force recalculation
                            inport.getNode().graph.itemMoved()

                            # trigger name update
                            inport.getNode().refreshName()
                            outport.getNode().refreshName()

                            # trigger event queue, if its idle
                            inport.getNode().graph._switchSig.emit('check')

                            if len(self.scene().portMatches):
                                for port in self.scene().portMatches:
                                    port.resetScale()
                                self.scene().portMatches = []

                            inport.edges()[0].adjust()
                            for edge in outport.edges():
                                edge.adjust()

                    connected.append(inport)

    def reload_node(self):
        '''Reload, instantiate, and reconnect the selected node.
                -Only allow one node.
        '''
        nodes = self.getSelectedNodes()

        if len(nodes) == 0:
            return

        # pause the canvas during this process
        alreadyPaused = self.inPausedState()
        if not alreadyPaused:
            self.pauseToggle(quiet=True)

        # copy
        self.copyNodesToBuffer()

        # delete node
        for node in nodes:
            self.deleteNode(node)

        # paste
        if self.parent._copybuffer:
            s = {'sig': 'load', 'subsig': 'reload'}
            self.addNodeRun(s)

        # unpause if the user hadn't already done so.
        if not alreadyPaused:
            self.pauseToggle()


    def closeAllNodeMenus(self):
        for node in self.getAllNodes():
            node.closemenu()

    def organizeSelectedNodes(self):
        nodes = self.getSelectedNodes()
        if len(nodes):
            snodes = self.getLinearNodeHierarchy_fromList(nodes)
            topnode = snodes[0]
            x = topnode.scenePos().x()
            y = topnode.scenePos().y()

            self._node_anim_group = QtCore.QParallelAnimationGroup()
            for node in snodes:
                anim = QtCore.QPropertyAnimation(node, b"pos")
                anim.setDuration(100)
                anim.setStartValue(QtCore.QPointF(x, topnode.scenePos().y()))
                anim.setEndValue(QtCore.QPointF(x, y))
                self._node_anim_group.addAnimation(anim)
                y += node.getNodeHeight() + 15.0

            self._node_anim_group.start()

    def chargeRepTimer(self, event):
        if self.chargeRepON is False:
            return

        nodes = self.getAllNodes()

        # don't let a single unattached node be affected by force
        nodes = [item for item in nodes if len(item.edges()) > 0]

        for node in nodes:
            node.calculateForces()

        itemsMoved = False
        for node in nodes:
            if node.advance():
                itemsMoved = True

        if not itemsMoved:
            self.killTimer(self.timerId)
            self.timerId = 0

    def timerEvent(self, event):
        # if event.timerId() == self.nodeEvent_timerId:
        #    print "timer: global node timer"
        # always update charge rep events
        self.chargeRepTimer(event)

    def wheelEvent(self, event):
        try:
            # PyQt4
            self.scaleView(math.pow(2.0, event.delta() / 300.0))
        except AttributeError:
            # PyQt5
            angle = event.angleDelta().y() / 8
            self.scaleView(math.pow(2.0, angle / 80.0))

    def drawBackground(self, painter, rect):
        # Shadow.
        sceneRect = self.sceneRect()
        rightShadow = QtCore.QRectF(sceneRect.right(), sceneRect.top() + 5, 5,
                                    sceneRect.height())
        bottomShadow = QtCore.QRectF(sceneRect.left() + 5, sceneRect.bottom(),
                                     sceneRect.width(), 5)
        if rightShadow.intersects(rect) or rightShadow.contains(rect):
            painter.fillRect(rightShadow, QtCore.Qt.darkGray)
        if bottomShadow.intersects(rect) or bottomShadow.contains(rect):
            painter.fillRect(bottomShadow, QtCore.Qt.darkGray)

        # Fill.
        gradient = QtGui.QLinearGradient(sceneRect.topLeft(),
                                         sceneRect.bottomRight())

        if self.inPausedState() and not self._pause_quiet:
            gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.yellow).lighter(190))
            gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.yellow).lighter(170))
        else:
            #gradient.setColorAt(0, QtCore.Qt.white)
            #gradient.setColorAt(1, QtCore.Qt.lightGray)
            gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.gray).lighter(180))
            gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.gray).lighter(150))

        painter.fillRect(rect.intersected(sceneRect), QtGui.QBrush(gradient))
        painter.setBrush(QtCore.Qt.NoBrush)
        painter.drawRect(sceneRect)

        # Text.
        textRect = QtCore.QRectF(sceneRect.left() + 4, sceneRect.top() + 4,
                                 sceneRect.width() - 4, sceneRect.height() - 4)
        message = "Network Canvas"

        font = painter.font()
        font.setBold(True)
        font.setPointSize(14)
        painter.setFont(font)
        painter.setPen(QtCore.Qt.lightGray)
        painter.drawText(textRect.translated(2, 2), message)
        painter.setPen(QtCore.Qt.black)
        painter.drawText(textRect, message)

        # Mark.
        ## centered
        #mark_font = QtGui.QFont(u"gill sans", 100)
        #fm = QtGui.QFontMetricsF(mark_font)
        #message = "PHILIPS"
        #bw = fm.width(message) * 1.12

        ## bw is the width at 100pt font
        #w = self.viewport().rect().width()
        #f = (100*w)/bw/2
        #mark_font = QtGui.QFont(u"gill sans", int(f))

        #mark_font = QtGui.QFont(u"gill sans", 50)
        #fm = QtGui.QFontMetricsF(mark_font)
        #bw = fm.width(message) * 1.12

        #bh = fm.height()
        # centered
        #textRect = QtCore.QRectF(self.mapToScene(self.viewport().rect().center()).x()-bw/2, self.mapToScene(self.viewport().rect().center()).y()-bh/2, bw, bh)
        #textRect = QtCore.QRectF(self.mapToScene(self.viewport().rect().bottomRight()).x()-bw-8, self.mapToScene(self.viewport().rect().bottomRight()).y()-bh-2, bw, bh)


        #mark_font.setBold(True)
        #mark_font.setPointSize(14)
        #painter.setFont(mark_font)
        #c = QtGui.QColor(QtCore.Qt.gray).lighter(130)
        #c.setAlphaF(0.5)
        #painter.setPen(c)
        #painter.drawText(textRect.translated(2, 2), message)
        #c = QtGui.QColor(QtCore.Qt.gray).lighter(150)
        #c.setAlphaF(0.5)
        #painter.setPen(c)
        #painter.setPen(QtCore.Qt.black)
        #painter.drawText(textRect, message)

    def scaleView(self, scaleFactor):
        try:
            # PyQt4
            factor = self.matrix().scale(scaleFactor, scaleFactor).mapRect(
                QtCore.QRectF(0, 0, 1, 1)).width()
        except AttributeError:
            # PyQt5 doesn't have a matrix() attribute
            factor = scaleFactor

        if factor < 0.07 or factor > 100:
            return

        self.scale(scaleFactor, scaleFactor)

    # def mouseDoubleClickEvent(self, event):
    #    event.accept()
    #    print "double-clicked canvas"

    def mousePressEvent(self, event):  # GRAPHICS VIEW
        printMouseEvent(self, event)
        modifiers = getKeyboardModifiers()

        self.viewAndSceneForcedUpdate()

        if self._panning:
            return

        QtWidgets.QGraphicsView.mousePressEvent(self, event)
        if event.isAccepted():
            return

        if event.button() == QtCore.Qt.MidButton:
            event.accept()
            self._panning = True
            # trick graphics view into thinking it has a left click for panning
            leftbutton_event = QtGui.QMouseEvent(
                event.type(), event.pos(), event.globalPos(),
                QtCore.Qt.LeftButton, event.buttons(), modifiers)
            leftbutton_event.accept()
            super(GraphWidget, self).mousePressEvent(leftbutton_event)
            return

        # propagate to other view items (nodes)
        if event.button() == QtCore.Qt.RightButton:
            event.accept()
            # if self.scene().itemAt(event.scenePos()):
            #    self.scene().unselectAllItems()
            #    self.scene().itemAt(event.scenePos()).setSelected(True)
            pointedItem = self.itemAt(event.pos())
            if not isinstance(pointedItem, InPort):
                self.rightButtonMenu(event)

        elif event.button() == QtCore.Qt.LeftButton:
            event.accept()
            # self.scene().unselectAllItems()
        else:
            event.ignore()
            QtWidgets.QGraphicsView.mousePressEvent(self, event)
            super(GraphWidget, self).mousePressEvent(event)

    def mouseMoveEvent(self, event):
        if self._panning or self.scene().rubberBand or self.scene().line:
            self.viewAndSceneForcedUpdate()
        self.mousePos = self.mapToScene(QtCore.QPoint(event.x(), event.y()))
        super(GraphWidget, self).mouseMoveEvent(event)

    def mouseReleaseEvent(self, event):  # GRAPHICS VIEW
        printMouseEvent(self, event)
        modifiers = getKeyboardModifiers()

        self.viewAndSceneForcedUpdate()

        if self._panning:
            event.accept()
            leftbutton_event = QtGui.QMouseEvent(
                event.type(), event.pos(), event.globalPos(),
                QtCore.Qt.LeftButton, event.buttons(), modifiers)
            leftbutton_event.accept()
            super(GraphWidget, self).mouseReleaseEvent(leftbutton_event)
            self._panning = False
            return

        # delete edge via input port
        if event.button() == QtCore.Qt.RightButton:
            pointedItem = self.itemAt(event.pos())
            if isinstance(pointedItem, InPort):
                edge = pointedItem.edge()
                if edge:
                    self.scene().removeItem(edge)
                    # remove from ports
                    edge.detachSelf(tracer=True)
                    event.accept()

        # propagate to other view items (nodes)
        QtWidgets.QGraphicsView.mouseReleaseEvent(self, event)
        if event.isAccepted():
            return

        if event.button() == QtCore.Qt.RightButton:
            pointedItem = self.itemAt(event.pos())
            if isinstance(pointedItem, InPort):
                event.accept()

        elif event.button() == QtCore.Qt.LeftButton:
            event.accept()
        # elif event.button() == QtCore.Qt.MidButton:
        #    event.accept()
        else:
            event.ignore()
            QtWidgets.QGraphicsView.mouseReleaseEvent(self, event)
            super(GraphWidget, self).mouseReleaseEvent(event)

    def rightButtonMenu(self, event):
        # MOUSE MENU
        pointedItem = self.itemAt(event.pos())
        if isinstance(pointedItem, Edge):
            event.accept()
            pointedItem.setSelected(True)
            pointedItem.update()
            menu = QtWidgets.QMenu(self)
            deleteEdgeAction = menu.addAction("Delete")
            action = menu.exec_(self.mapToGlobal(event.pos()))
            if action == deleteEdgeAction:
                # remove from scene
                self.scene().removeItem(pointedItem)
                # remove from ports
                pointedItem.detachSelf()
                # remove from memory
                # del pointedItem
            else:
                pointedItem.setSelected(False)
                pointedItem.update()
        else:
            event.accept()

            # save position before any choice is made
            self._event_pos = event.pos()

            # main menu
            menu = QtWidgets.QMenu(self.parent)

            # search
            qle = QtWidgets.QLineEdit()
            qle.setContextMenuPolicy(QtCore.Qt.NoContextMenu)
            qle.setPlaceholderText('  Search')
            qle.textChanged.connect(lambda txt: self._library.searchMenu(txt, qle, menu))
            wac = QtWidgets.QWidgetAction(menu)
            msg = 'Search for nodes and networks in the library.'
            wac.hovered.connect(lambda who=msg: self.setStatusTip(who))
            wac.setDefaultWidget(qle)
            menu.addAction(wac)
            menu.addSeparator()

            # favorites at the top, then full library
            #if 'Favorites' in self._library.libMenus():
            #    menu.addMenu(self._library.libMenus()['Favorites'])

            # add a text label
            #menu.addSeparator()
            #menu.addAction(u'\u25BC'+u' Libraries')

            menu.addSeparator()

            for libmenu in self._library.libMenu():
                ma = menu.addMenu(libmenu) # previously generated
                msg = 'Select nodes from the \''+str(ma.text())+'\' library.'
                ma.hovered.connect(lambda who=msg: self.setStatusTip(who))

            pasteAct = QtWidgets.QAction("&Paste", self, shortcut="Ctrl+V",
                            statusTip="Paste node(s) from the copybuffer.")

            copyAct = QtWidgets.QAction("Copy", self, shortcut="Ctrl+C",
                            statusTip="Copy node(s) to the copybuffer.")

            saveAct = QtWidgets.QAction("Save Network", self, shortcut="Ctrl+S",
                            statusTip="Save network to a file.")

            loadAct = QtWidgets.QAction("Load Network", self, shortcut="Ctrl+L",
                            statusTip="Load network from a file (also drag'n drop).")

            layoutMenu = QtWidgets.QMenu('New Layout')
            layoutMenu.addAction(QtWidgets.QAction("Vertical", self,
                            statusTip="Opens a vertically expanding layout window for widgets.",
                            triggered = lambda: self.newLayoutWindow(config=0)))
            layoutMenu.addAction(QtWidgets.QAction("Horizontal", self,
                            statusTip="Opens a horizontally expanding layout window for widgets.",
                            triggered = lambda: self.newLayoutWindow(config=1)))
            layoutMenu.addAction(QtWidgets.QAction("Mixed", self,
                            statusTip="Opens a fixed & mixed layout window for widgets.",
                            triggered = lambda: self.newLayoutWindow(config=2)))
            layoutMenu.addAction(QtWidgets.QAction("Expanding", self,
                            statusTip="Opens an expanding mixed layout window for widgets.",
                            triggered = lambda: self.newLayoutWindow(config=3)))

            quitAct = QtWidgets.QAction("Quit", self,
                            statusTip="Quit GPI without saving.")

            clearAct = QtWidgets.QAction("Clear Canvas", self,
                            statusTip="Delete all nodes on the canvas.")

            pauseAct = QtWidgets.QAction("Pause", self, shortcut="Ctrl+P", checkable = True, triggered=lambda: self.pauseToggle(quiet=False),
                            statusTip="Pause the execution queue on this canvas.")
            if self.inPausedState():
                pauseAct.setChecked(True)
            else:
                pauseAct.setChecked(False)

            macroAct = QtWidgets.QAction("Macro Node", self, checkable = True, triggered = lambda: self.newMacroNode(event.pos()),
                            statusTip="Instantiate Macro Node Objects.")
            if self.isMacroModule():
                macroAct.setChecked(True)
            else:
                macroAct.setChecked(False)

            # basic editor actions
            menu.addSeparator()
            menu.addAction(copyAct)
            menu.addAction(pasteAct)
            menu.addSeparator()
            menu.addAction(saveAct)
            menu.addAction(loadAct)
            menu.addSeparator()
            menu.addAction(pauseAct)
            menu.addSeparator()
            menu.addAction(clearAct)
            menu.addAction(macroAct)
            menu.addMenu(layoutMenu)
            #menu.addSeparator()
            
            # trigger a search menu close when the main menu is hovered
            menu.hovered.connect(lambda: self._library.removeSearchPopup())

            #quitAction = menu.addAction(quitAct)
            action = menu.exec_(self.mapToGlobal(event.pos()))

            self._library.removeSearchPopup()

            if action == copyAct:
                self.copyNodesToBuffer()

            if action == pasteAct:
                # mouse-menu paste
                s = {'sig': 'load', 'subsig': 'paste', 'pos': self.mapToScene(event.pos())}
                self.addNodeRun(s)

            if action == clearAct:  # DELETE
                #self._switchSig.emit('deleteAll')  # change state

                reply = QtWidgets.QMessageBox.question(self, 'Message',
                            "Delete all modules on this canvas?", QtWidgets.QMessageBox.Yes |
                                QtWidgets.QMessageBox.No, QtWidgets.QMessageBox.No)

                if reply == QtWidgets.QMessageBox.Yes:
                    self.deleteNodeRun('deleteAll')

            if action == loadAct:
                s = {'sig': 'load', 'subsig': 'dialog',
                     'pos': self.mapToScene(event.pos())}
                self.addNodeRun(s)
                #self._switchSig_info.emit(s)

            if action == saveAct:
                self._network.saveNetworkFromFileDialog(self.serializeCanvas())

            if action == quitAct:
                pass
                # pausing might make quitting more graceful
                #self._switchSig.emit('pause')
                #self.closeGraph(event)
                #QtWidgets.qApp.quit()

            self.parent.statusBar().clearMessage()

    def setStatusTip(self, msg):
        self.parent.statusBar().showMessage(msg)

    def newMacroNode(self, pos):
        log.debug("drop new macro")
        log.debug(str(pos))
        if isinstance(pos, QtCore.QPointF):
            pos = QtCore.QPoint(pos.x(), pos.y())
        newnode = MacroNode(self, QtCore.QPointF(self.mapToScene(pos)))

        #self.scene().addItem(newnode)
        return newnode


    def showMacroTools(self):
        '''Show the src, sink, and macro layout window.
        '''
        self.macroModuleToggle()

        if self.isMacroModule():
            log.debug("show macro tools")
        else:
            reply = QtWidgets.QMessageBox.question(self, 'Message',
                    "Turn off macro settings for this canvas?\n\nAny src/sink " + \
                    "connections and macro-layouts will be removed.",
                    QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No,
                    QtWidgets.QMessageBox.No)

            if reply == QtWidgets.QMessageBox.Yes:
                log.debug("hide macro tools")
            else:
                log.debug("cancel hide")
                self.setMacroModule(True)

    def isMacroModule(self):
        return self._macroModule

    def setMacroModule(self, val):
        self._macroModule = val

    def macroModuleToggle(self):
        if self._macroModule:
            self._macroModule = False
        else:
            self._macroModule = True

    def closeEvent(self, event):
        self.closeGraphNoDialog()
        event.accept()

    def closeGraphWithDialog(self):
        reply = QtWidgets.QMessageBox.question(self, 'Message',
                    "Close canvas without saving?", QtWidgets.QMessageBox.Yes |
                        QtWidgets.QMessageBox.No, QtWidgets.QMessageBox.No)

        if reply == QtWidgets.QMessageBox.Yes:
            self.closeGraphNoDialog()
            return True
        return False

    def closeGraphNoDialog(self):
        '''This is the close procedure for a canvas.
        '''
        self._switchSig.emit('pause')
        self.deleteNodeRun('deleteAll')
        for lm in self._layoutwindowList:
            if lm:  # some may have already been closed
                lm.close()

    def pauseToggle(self, quiet=False):
        '''toggle the pause state for attached buttons.
        '''
        if self.inPausedState():
            self._switchSig.emit('unpause')
        else:
            if quiet:
                self._switchSig_info.emit({'sig':'pause', 'subsig':'quiet'})
            else:
                self._switchSig.emit('pause')

    def copyNodesToBuffer(self):
        self.parent._copybuffer = self.serializeGraphData(selectedOnly=True)

    # NETWORK SERIALIZATION

    def deserializeGraphData(self, graph_settings, layoutSettings=[], pos=None, offset=False, randoffset=False, reloadnode=False):
        log.info("num nodes: " + str(len(graph_settings['nodes'])))

        # determine network center randomly to avoid overlap
        rx = 0
        ry = 0
        if offset:
            # this places modules at a slight offset to their orig-pos.
            # used for keystroke copy/paste
            # -keep original position and add this offset
            radius = 5.0  # pts
            rx += radius * 6.0
            ry += radius * 6.0

        if randoffset:
            # if a network is loaded by menu more than once, this small
            # perturbation makes it easy to distinguish between multiple
            # instances
            radius = 5.0  # pts
            rx += random.random() * radius
            ry += random.random() * radius

        if pos:
            # if the position is supplied then the graph should be
            # instantiated relative to it.
            graph_settings = self.subtractAvgPosFromSettings(graph_settings)
            rx += pos.x()
            ry += pos.y()

        # temporarily buffer nodes in case this network is going to be merged
        buf = []
        macro_buf = []
        skipped_mods = []
        new_nodes = []

        # for reloading nodes use the existing nodes on the canvas so that
        # the IDs match for connecting edges.
        if reloadnode:
            buf += self.getAllNodes()

        # place all nodes on the canvas
        for s in graph_settings['nodes']:

            log.debug("add node: " + str(s['name']))

            # try to import the node module by name
            if s['name'] == '__GPIMacroNode__':
                continue

            # instantiate node
            cpos = QtCore.QPoint(s['pos'][0] + rx, s['pos'][1] + ry)

            # first always try to get the node by library
            node = self.newNode_byKey(s['key'], cpos)
            if node is None:
                log.warn('Failed to find node \''+stw(s['name']) + '\' by scope.')

                # get a list of widget and port names together
                wdg_port_names = []
                for parm in s['widget_settings']['parms']:
                    wdg_port_names.append(parm['name'])
                for port in s['ports']:
                    wdg_port_names.append(port['porttitle'])

                # find from libarary and instantiate on the canvas
                node = self.newNode_byClosestMatch(s['name'], wdg_port_names, cpos)

            # final failure to resolve node
            if node is None:
                log.error('Node \''+stw(s['name']) + '\' failed to load, skipping.')
                skipped_mods.append(str(s['name']))
                continue

            new_nodes.append(node)
            node.setDisabledState(True)  # put in disabled state
            node.setID(s['id'])
            buf.append(node)

            # set other node attributes
            if 'walltime' in s:
                try:  # deprecate this try statement
                    node.appendWallTime(float(s['walltime']))
                except:
                    log.error(stw(node.getModuleName()) + ' has no walltime but walltime was saved as NoneType, skipping...')

            node.loadNodeIFSettings(s['widget_settings'])

        # place all macro nodes on the canvas and load settings
        #   -done after node instantiation so that widgets can be copied over
        log.debug("Load MacroNodes:")
        log.debug(str(graph_settings['macroNodes']))
        for s in graph_settings['macroNodes']:
            mnode = MacroNode(self, QtCore.QPointF(rx, ry))
            mnode.loadSettings(s, buf, (rx, ry))
            macro_buf.append(mnode)

            for node in mnode.getNodes():
                buf.append(node)

        # once all nodes have been placed,
        # start making connections
        for s in graph_settings['nodes']:

            # remake the connections
            for port in s['ports']:
                for c in port['connections']:
                    # make a new edge given src and dest
                    # get the nodes
                    src = self.getNodeByID(buf, c['src']['nodeID'])
                    dst = self.getNodeByID(buf, c['dest']['nodeID'])

                    if src and dst:
                        # get the ports
                        #outport = src.getPortByNumOrTitle(c['src']['portName'])
                        #inport = dst.getPortByNumOrTitle(c['dest']['portName'])
                        outport = src.getOutPort(c['src']['portName'])
                        inport = dst.getInPort(c['dest']['portName'])

                        # each connection is stored twice (memory of each node)
                        # only use one for each pair
                        try:
                            log.debug("inport title: "+inport.portTitle)
                            if len(inport.edges()) > 0:
                                log.debug("Inport occupied," \
                                    + " connection dropped.")
                            else:
                                # make the connection
                                print("here cg")
                                newEdge = Edge(outport, inport)
                                self.scene().addItem(newEdge)
                        except:
                            log.warn("Duplicate or connection" \
                                + " not found.  Skip connection.")
                    else:
                        log.warn("Src and Dst not" \
                            + " accurately saved.  Skip connection.")

        self.scene().unselectAllItems()

        # load layouts before the ids get reset.
        for lw in layoutSettings:
            self.newLayoutWindowFromSettings(lw, buf)

        # reset node IDs, and widget IDs
        for node in buf:
            node.setID()
            for parm in node.getParmList():
                parm.set_id()

        # reset node IDs, and widget IDs select only newly loaded network items
        if reloadnode:
            # for reloading nodes
            for node in new_nodes:
                node.setSelected(True)
                node.setEventStatus({GPI_INIT_EVENT: None})
                node.displayReloaded()
        else:
            # for importing networks
            for node in buf:
                node.setSelected(True)
                node.setEventStatus({GPI_INIT_EVENT: None})

        # reset macro IDs
        for node in macro_buf:
            if node.shouldCollapse():
                node.setCollapse(True)
            node.resetIDs()

        self.calcNodeHierarchy()

        # put nodes back in to idle
        for node in buf:
            node.setDisabledState(False)

        QtWidgets.QApplication.processEvents()  # allow gui to update

        if len(skipped_mods):
            log.error("Failed to load the following modules: ")
            for name in skipped_mods:
                log.error("\t" + name)

        if not reloadnode:
            try:
                # get top most node position-wise
                # and make sure its visible
                topnode = buf[0]
                for node in buf:
                    if node.pos().y() < topnode.pos().y():
                        topnode = node
                self.ensureVisible(topnode)
            except:
                log.warn("Can\'t determine top node, skipping.")

    def getNodeByID(self, buf, nid):
        for item in buf:
            if isinstance(item, Node):
                if item.getID() == nid:
                    return item

    def calcAvgPosFromSettings(self, graph_settings):
        cnt = 0.
        ax = 0.
        ay = 0.

        # macro nodes
        for mnode in graph_settings['macroNodes']:
            node = mnode['src_settings']
            ax += node['pos'][0]
            ay += node['pos'][1]
            cnt += 1.

            node = mnode['sink_settings']
            ax += node['pos'][0]
            ay += node['pos'][1]
            cnt += 1.

            node = mnode['face_settings']
            ax += node['pos'][0]
            ay += node['pos'][1]
            cnt += 1.

        # normal nodes
        for node in graph_settings['nodes']:
            ax += node['pos'][0]
            ay += node['pos'][1]
            cnt += 1.

        ax = ax / cnt
        ay = ay / cnt
        return(ax, ay)

    def subtractAvgPosFromSettings(self, graph_settings):
        ax, ay = self.calcAvgPosFromSettings(graph_settings)
        newgraphsettings = copy.deepcopy(graph_settings)
        #newgraphsettings = graph_settings

        # macro nodes
        for mnode in newgraphsettings['macroNodes']:
            node = mnode['src_settings']
            node['pos'][0] -= ax
            node['pos'][1] -= ay

            node = mnode['sink_settings']
            node['pos'][0] -= ax
            node['pos'][1] -= ay

            node = mnode['face_settings']
            node['pos'][0] -= ax
            node['pos'][1] -= ay

        for node in newgraphsettings['nodes']:
            node['pos'][0] -= ax
            node['pos'][1] -= ay

        return(newgraphsettings)

    def serializeGraphData(self, selectedOnly=False, minusAvgPos=False):
        # Handles only nodes and macro-nodes.  The 'selectedOnly' option is
        # for copy/paste operation.

        graph_settings = {}
        graph_settings['nodes'] = []
        graph_settings['macroNodes'] = []

        # serialize all nodes and macro nodes
        if selectedOnly:
            nodes = self.getSelectedNodes()
            macroNodes, enodes = self.getSelectedMacroNodes()
            nodes = list(set(nodes + enodes))
        else:
            nodes = self.getAllNodes()
            macroNodes, enodes = self.getAllMacroNodes()
            nodes = list(set(nodes + enodes))

        for node in nodes:
            graph_settings['nodes'].append(copy.deepcopy(node.getSettings()))

        for nid, nodes in list(macroNodes.items()):
            graph_settings['macroNodes'].append(nodes[0].macroParent().getSettings())

        if minusAvgPos and (len(graph_settings['nodes']) + len(graph_settings['macroNodes'])):
            graph_settings = self.subtractAvgPosFromSettings(graph_settings)

        return graph_settings

    def serializeCanvas(self):
        network = {}
        network['nodes'] = self.serializeGraphData(minusAvgPos=True)
        network['layouts'] = self.serializeLayoutWindows()
        network['WALLTIME'] = str(self.walltime())  # sec
        network['TOTAL_PMEM'] = str(self.totalPortMem())  # bytes
        return network

    def deserializeCanvas(self, network, pos):
        # convert the loaded file data into the format required by GPI objects
        # and instantiate the network on the canvas.
        nodes = network['nodes']
        layouts = network['layouts']

        if nodes:
            log.info("load nodes.")
        else:
            log.error("network description contains no node information!!!")
            return

        if layouts:
            log.info("network has layouts.")
            self.deserializeGraphData(nodes, layoutSettings=layouts, pos=pos)
        else:
            self.deserializeGraphData(nodes, pos=pos)





#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

# Brief: tee stdout and stderr to a built-in console.
# TODO: 1) find a way to get stdout from c-pymods.
#       2) possibly integrate qconsole or similar terminal emulator to launch gpi


import sys

# gpi
import gpi
from gpi import QtCore

## rediect objects for stdout

class StreamBuf(object):

    def __init__(self, stdout):
        self._stdout = stdout
    
    def write(self, msg):
        self._stdout.write('gpi: '+msg)

    def flush(self):
        self._stdout.flush()

# support functions
# http://shallowsky.com/blog/programming/python-tee.html
class Tee(QtCore.QObject):
    '''(Mostly Unused) An attempt at providing console output to a GPI internal
    console window.  This is part of an unfinished feature that is meant to
    give the user easy access to logging information. '''

    newStreamTxt = gpi.Signal(str)

    def __init__(self, stdIO=None, parent=None):
        """Redirect to a pyqtSignal and stdIO stream.
        stdIO = alternate stream ( can be the original sys.stdout )
        """
        super(Tee, self).__init__(parent)
        self._stdIO = stdIO
        self._fromProc = False

    def setMultiProc(self, val=True):
        self._fromProc = val
    def isMultiProc(self):
        return self._fromProc

    def write(self, m):
        #if self._color and self._edit:
        #    tc = self._edit.textColor()
        #    self._edit.setTextColor(self._color)
#
#        if self._edit:
#            self._edit.moveCursor(QtGui.QTextCursor.End)
#            self._edit.insertPlainText( m )
#
#        if self._color and self._edit:
#            self._edit.setTextColor(tc)
#
        if not self.isMultiProc():
            self.newStreamTxt.emit(m)

        if self._stdIO:
            self._stdIO.write(m)

    def flush(self):
        self._stdIO.flush()



# tee both stdout and stderr for the console
#sys.stdout = StreamBuf(sys.stdout)
#sys.stderr = sys.stdout



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: gpi.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='gpi.proto',
  package='',
  syntax='proto3',
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_pb=b'\n\tgpi.proto\")\n\x07Request\x12\x10\n\x08\x66unction\x18\x01 \x01(\x0c\x12\x0c\n\x04\x61rgs\x18\x02 \x01(\x0c\"\x18\n\x06Result\x12\x0e\n\x06result\x18\x01 \x01(\x0c\x32!\n\x03GPI\x12\x1a\n\x03run\x12\x08.Request\x1a\x07.Result\"\x00\x62\x06proto3'
)




_REQUEST = _descriptor.Descriptor(
  name='Request',
  full_name='Request',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='function', full_name='Request.function', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
    _descriptor.FieldDescriptor(
      name='args', full_name='Request.args', index=1,
      number=2, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=13,
  serialized_end=54,
)


_RESULT = _descriptor.Descriptor(
  name='Result',
  full_name='Result',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  create_key=_descriptor._internal_create_key,
  fields=[
    _descriptor.FieldDescriptor(
      name='result', full_name='Result.result', index=0,
      number=1, type=12, cpp_type=9, label=1,
      has_default_value=False, default_value=b"",
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR,  create_key=_descriptor._internal_create_key),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=56,
  serialized_end=80,
)

DESCRIPTOR.message_types_by_name['Request'] = _REQUEST
DESCRIPTOR.message_types_by_name['Result'] = _RESULT
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

Request = _reflection.GeneratedProtocolMessageType('Request', (_message.Message,), {
  'DESCRIPTOR' : _REQUEST,
  '__module__' : 'gpi_pb2'
  # @@protoc_insertion_point(class_scope:Request)
  })
_sym_db.RegisterMessage(Request)

Result = _reflection.GeneratedProtocolMessageType('Result', (_message.Message,), {
  'DESCRIPTOR' : _RESULT,
  '__module__' : 'gpi_pb2'
  # @@protoc_insertion_point(class_scope:Result)
  })
_sym_db.RegisterMessage(Result)



_GPI = _descriptor.ServiceDescriptor(
  name='GPI',
  full_name='GPI',
  file=DESCRIPTOR,
  index=0,
  serialized_options=None,
  create_key=_descriptor._internal_create_key,
  serialized_start=82,
  serialized_end=115,
  methods=[
  _descriptor.MethodDescriptor(
    name='run',
    full_name='GPI.run',
    index=0,
    containing_service=None,
    input_type=_REQUEST,
    output_type=_RESULT,
    serialized_options=None,
    create_key=_descriptor._internal_create_key,
  ),
])
_sym_db.RegisterServiceDescriptor(_GPI)

DESCRIPTOR.services_by_name['GPI'] = _GPI

# @@protoc_insertion_point(module_scope)



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.


import math
from gpi import QtCore, QtGui, QtWidgets

# gpi
from .defaultTypes import GPITYPE_PASS
from .defines import MacroNodeEdgeType, EdgeNodeType, PortEdgeType, GPI_APPLOOP
from .defines import printMouseEvent, getKeyboardModifiers, OPTIONAL
from .defines import isMacroChildNode
from .layoutWindow import LayoutMaster
from .logger import manager
from .nodeAPI import NodeAPI
from .node import Node, node_font

# start logger for this module
log = manager.getLogger(__name__)


class EdgeNode(QtWidgets.QGraphicsObject, QtWidgets.QGraphicsItem):
    '''The EdgeNode simply forwards port connections.  It can take one OutPort
    connection, and multiple InPort connections. InPort tooltips and
    enforcement are concatenated and forwarded.
    '''

    Type = EdgeNodeType

    def __init__(self, portNum, parentItem):
        super(EdgeNode, self).__init__()

        self.setParentItem(parentItem)

        # position
        self.portNum = portNum

        # attached ports
        self.portList = []

    def type(self):
        return self.Type


class MacroAPI(NodeAPI):
    '''A class that has default operations that stub-out the NodeAPI.
    '''

    def initUI(self):
        pass

        # self._layoutWindow = canvasGraph.LayoutMaster(self.node.graph, 0)
        # self.layout.addWidget(self._layoutWindow, 0, 0)

    def execType(self):
        return GPI_APPLOOP

    def validate(self):
        '''Copy all inputs to outputs.
        '''
        return 0

    def compute(self):
        '''Copy all inputs to outputs.
        '''
        for i in range(self.node._numPorts):
            self.setData('out' + str(i + 1), self.getData('in' + str(i + 1)))
        return 0


class PortEdge(Node):
    '''The PortEdge holds EdgeNodes for the MacroNode (they are the two node
    like pieces that bound the macro and the single collapsed macro
    representation).
    '''

    Type = PortEdgeType

    def __init__(self, macroParent, graph, menu=None, nodeIF=None, nodeIFscroll=None, role=None):
        super(PortEdge, self).__init__(graph, nodeMenuClass=menu, nodeCatItem=None, nodeIF=nodeIF, nodeIFscroll=nodeIFscroll)

        self._roles = ['Input', 'Output', 'Macro']
        self._role = self._roles[role]
        self.name = self._role
        self._label = ''
        self._title_delimiter = ': '
        self._macro_prefix = '*'

        self._macroParent = macroParent
        self._isMacroNode = False
        if self._macroParent:
            self._isMacroNode = True

        # code name for network loader to ignore
        self._moduleName = '__GPIMacroNode__'

        self._macroEdges = []

        # Ports
        if self._role == 'Macro':
            self._numPorts = 0
        else:
            self._numPorts = 16

        for i in range(self._numPorts):
            self.addInPort('in' + str(i + 1), GPITYPE_PASS, obligation=OPTIONAL)
            self.addOutPort('out' + str(i + 1), GPITYPE_PASS)

        self._nodeIF.setTitle(self.name)

    def menu(self):

        self._macroParent._scrollArea_layoutWindow.show()
        self._macroParent._scrollArea_layoutWindow.raise_()
        # self._macroParent._scrollArea_layoutWindow.activateWindow()

        # self._macroParent._layoutWindow.show()
        # self._macroParent._layoutWindow.raise_()
        self._macroParent._layoutWindow.activateWindow()

    def closeMacroLayout(self):
        if hasattr(self._nodeIF, '_layoutWindow'):  # sibling safe
            log.debug("Macro Node issuing a force close on layout")
            self._nodeIF._layoutWindow.forceClose()
            # return self._nodeIF._layoutWindow.close()
        # return True

    def getSiblingNodes(self):
        return self._macroParent.getNodes()

    def macroParent(self):
        return self._macroParent

    def itemChange(self, change, value):
        if change == QtWidgets.QGraphicsItem.ItemPositionHasChanged:
            for edge in self._macroEdges:
                edge.adjust()
        return super(PortEdge, self).itemChange(change, value)

    def type(self):
        return self.Type

    def addMacroEdge(self, edge):
        self._macroEdges.append(edge)
        edge.adjust()

    def mousePressEvent(self, event):
        printMouseEvent(self, event)
        modifiers = getKeyboardModifiers()

        self.update()  # update node color
        modmidbutton_event = (event.button() == QtCore.Qt.LeftButton
                              and modifiers == QtCore.Qt.AltModifier)
        if event.button() == QtCore.Qt.MidButton or modmidbutton_event:
            event.accept()
        elif event.button() == QtCore.Qt.LeftButton:
            event.accept()  # this has to accept to be moved
        elif event.button() == QtCore.Qt.RightButton:
            event.accept()
        else:
            event.ignore()
        super(PortEdge, self).mousePressEvent(event)

    def mouseReleaseEvent(self, event):
        printMouseEvent(self, event)
        modifiers = getKeyboardModifiers()

        self.update()
        modmidbutton_event = (event.button() == QtCore.Qt.LeftButton
                              and modifiers == QtCore.Qt.AltModifier)
        if event.button() == QtCore.Qt.MidButton or modmidbutton_event:
            event.accept()
        elif event.button() == QtCore.Qt.LeftButton:
            event.accept()  # this has to accept to be moved
        elif event.button() == QtCore.Qt.RightButton:
            event.accept()
            # self.menu()
            # self.graph.scene().makeOnlyTheseNodesSelected([self])
        else:
            event.ignore()
        super(PortEdge, self).mouseReleaseEvent(event)

    def mouseDoubleClickEvent(self, event):
        event.accept()
        log.debug("double-clicked PortEdge")
        self._macroParent.toggleCollapse()

    def getTitleWidth(self):
        '''Determine how long the module box is.'''
        buf = self.getMacroNodeName()
        fm = QtGui.QFontMetricsF(self.title_font)
        bw = fm.width(buf) + 11.0
        bh = fm.height()
        return (bw, bh)

    def getRoleTitleSize(self):
        buf = self._role
        fm = QtGui.QFontMetricsF(self.title_font)
        bw = fm.width(buf)
        bh = fm.height()
        return (bw, bh)

    def getTitleDelimiterSize(self):
        buf = self._title_delimiter
        fm = QtGui.QFontMetricsF(self.title_font)
        bw = fm.width(buf)
        bh = fm.height()
        return (bw, bh)

    def getNodeWidth(self):
        return max(self.getMaxPortWidth(), self.getTitleWidth()[0])

    def getMacroNodeName(self):
        '''The name is based on the Macro Node's role within the macro
        framework.
        '''
        buf = self._role
        if self._role == 'Input':
            if self._macroParent._label != '':
                buf += self._title_delimiter + self._macroParent._label
        elif self._role == 'Output':
            if self._macroParent._label != '':
                buf += self._title_delimiter + self._macroParent._label
        elif self._role == 'Macro':
            if self._macroParent._label != '':
                buf = self._macro_prefix + self._macroParent._label

        return buf

    def paint(self, painter, option, widget):  # NODE

        w = self.getNodeWidth()
        # h = self.getTitleWidth()[1]

        # draw shadow
        painter.setPen(QtCore.Qt.NoPen)
        painter.setBrush(QtCore.Qt.darkGray)
        painter.drawRoundedRect(-8, -8, w, 20, 3, 3)

        # choose module color
        gradient = QtGui.QRadialGradient(-10, -10, 40)

        # update the face node based on the state of internal nodes.
        if self._role == 'Macro':
            if self._macroParent.isProcessing():
                gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.gray).lighter(70))
                gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.darkGray).lighter(70))
            elif (option.state & QtWidgets.QStyle.State_Sunken) or (self._macroParent.inComputeErrorState()):
                gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.red).lighter(150))
                gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.red).lighter(170))
            elif self._macroParent.inValidateErrorState():
                gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.yellow).lighter(190))
                gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.yellow).lighter(170))
            elif self._macroParent.inInitUIErrorState():
                gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.red).lighter(150))
                gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.yellow).lighter(170))
            else:
                gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.gray).lighter(150))
                gradient.setColorAt(
                    1, QtGui.QColor(QtCore.Qt.darkGray).lighter(150))

        # let the src and sink nodes update themselves normally
        else:
            conf = self.getCurState()
            if self._computeState is conf:
                gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.gray).lighter(70))
                gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.darkGray).lighter(70))
            elif (option.state & QtWidgets.QStyle.State_Sunken) or (self._computeErrorState is conf):
                gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.red).lighter(150))
                gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.red).lighter(170))
            elif self._validateError is conf:
                gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.yellow).lighter(190))
                gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.yellow).lighter(170))
            else:
                gradient.setColorAt(0, QtGui.QColor(QtCore.Qt.gray).lighter(150))
                gradient.setColorAt(1, QtGui.QColor(QtCore.Qt.darkGray).lighter(150))

        # draw module box (apply color)
        painter.setBrush(QtGui.QBrush(gradient))
        if self.beingHovered or self.isSelected():
            #painter.setPen(QtGui.QPen(QtCore.Qt.red, 1))
            fade = QtGui.QColor(QtCore.Qt.red)
            fade.setAlpha(100)
            painter.setPen(QtGui.QPen(fade, 2))
        else:
            #painter.setPen(QtGui.QPen(QtCore.Qt.black, 0))
            fade = QtGui.QColor(QtCore.Qt.black)
            fade.setAlpha(50)
            painter.setPen(QtGui.QPen(fade,0))

        painter.drawRoundedRect(-10, -10, w, 20, 3, 3)

        # title
        painter.setPen(QtGui.QPen(QtCore.Qt.black, 0))
        painter.setFont(self.title_font)

        buf = self.getMacroNodeName()

        # paint the node title
        painter.drawText(-5, -9, w, 20, (QtCore.Qt.AlignLeft |
                         QtCore.Qt.AlignVCenter), str(buf))


class MacroNodeEdge(QtWidgets.QGraphicsObject, QtWidgets.QGraphicsItem):
    """Provides the connection graphic and logic for nodes.
    -No enforcement, just methods to retrieve connected nodes.
    """
    Type = MacroNodeEdgeType
    Pi = math.pi
    TwoPi = 2.0 * Pi

    def __init__(self, source, dest):
        super(MacroNodeEdge, self).__init__()

        self.setZValue(0)

        self.source = source  # PortEdge(pos, self)
        self.dest = dest  # PortEdge(QtCore.QPointF(0, 30)+pos, self)

        self.arrowSize = 20.0
        self.penWidth = 10.0
        self.sourcePoint = QtCore.QPointF()
        self.destPoint = QtCore.QPointF()

        self.setFlag(QtWidgets.QGraphicsItem.ItemSendsGeometryChanges)
        self.setCacheMode(QtWidgets.QGraphicsItem.DeviceCoordinateCache)

        self.setAcceptedMouseButtons(QtCore.Qt.NoButton)
        self.source.addMacroEdge(self)
        self.dest.addMacroEdge(self)
        self.adjust()

        self.fontHeight = 20

    def itemChange(self, change, value):
        if change == QtWidgets.QGraphicsItem.ItemPositionHasChanged:
            # call adjustments or updates to position here
            pass
        return super(MacroNodeEdge, self).itemChange(change, value)

    def type(self):
        return self.Type

    def sourceNode(self):
        return self.source

    def setSourceNode(self, node):
        self.source = node
        self.adjust()

    def destNode(self):
        return self.dest

    def setDestNode(self, node):
        self.dest = node
        self.adjust()

    def adjust(self):
        if not self.source or not self.dest:
            return

        # line = QtCore.QLineF(self.mapFromItem(self.source, 0, 0)+self.source.line().p1(),
        # self.mapFromItem(self.dest, 0, 0)+self.dest.line().p1())
        line = QtCore.QLineF(self.mapFromItem(self.source, 0, 0),
                             self.mapFromItem(self.dest, 0, 0))

        self.prepareGeometryChange()

        self.sourcePoint = line.p1()
        self.destPoint = line.p2()

    def boundingRect(self):
        if not self.source or not self.dest:
            return QtCore.QRectF()

        extra = (self.penWidth + self.arrowSize) / 2.0
        extra = max(extra, self.fontHeight)

        return QtCore.QRectF(self.sourcePoint,
                             QtCore.QSizeF(
                             self.destPoint.x() - self.sourcePoint.x(),
                             self.destPoint.y() - self.sourcePoint.y())).normalized().adjusted(-extra, -extra, extra, extra)

    def paint(self, painter, option, widget):
        if not self.source or not self.dest:
            return

        # Draw the line itself.
        line = QtCore.QLineF(self.sourcePoint, self.destPoint)

        if line.length() == 0.0:
            return

        painter.setPen(
            QtGui.QPen(QtCore.Qt.gray, self.penWidth, QtCore.Qt.SolidLine,
                       QtCore.Qt.RoundCap, QtCore.Qt.RoundJoin))
        painter.drawLine(line)

        # drawing text
        x = (line.x1() + line.x2()) / 2.0
        y = (line.y1() + line.y2()) / 2.0
        xa = (line.x1() - line.x2())
        ya = (line.y1() - line.y2())
        m = math.sqrt(xa * xa + ya * ya)
        a = math.atan2(ya, xa) * 180.0 / math.pi
        buf = "Macro"
        f = QtGui.QFont(node_font, 20)
        fm = QtGui.QFontMetricsF(f)
        bw = fm.width(buf)
        bw2 = -bw * 0.5
        # bh = fm.height()

        # Draw the arrows if there's enough room.
        angle = math.acos(line.dx() / line.length())
        if line.dy() >= 0:
            angle = self.TwoPi - angle

        sourceArrowP1 = self.sourcePoint + QtCore.QPointF(
            math.sin(angle + self.Pi / 3) * self.arrowSize,
            math.cos(angle + self.Pi / 3) * self.arrowSize)
        sourceArrowP2 = self.sourcePoint + QtCore.QPointF(
            math.sin(angle + self.Pi - self.Pi / 3) * self.arrowSize,
            math.cos(angle + self.Pi - self.Pi / 3) * self.arrowSize)
        destArrowP1 = self.destPoint + QtCore.QPointF(
            math.sin(angle - self.Pi / 3) * self.arrowSize,
            math.cos(angle - self.Pi / 3) * self.arrowSize)
        destArrowP2 = self.destPoint + QtCore.QPointF(
            math.sin(angle - self.Pi + self.Pi / 3) * self.arrowSize,
            math.cos(angle - self.Pi + self.Pi / 3) * self.arrowSize)

        painter.setBrush(QtCore.Qt.gray)
        painter.drawPolygon(
            QtGui.QPolygonF([line.p1(), sourceArrowP1, sourceArrowP2]))
        painter.drawPolygon(
            QtGui.QPolygonF([line.p2(), destArrowP1, destArrowP2]))

        # drawing text
        painter.setFont(f)
        painter.setPen(QtGui.QPen(QtCore.Qt.darkGray, 1))
        painter.save()
        painter.translate(QtCore.QPointF(x, y))
        if m > bw * 1.1:
            if a > 90 or a < -90:
                painter.rotate(a + 180.0)
                painter.drawText(QtCore.QPointF(bw2, -5.0), buf)
            else:
                painter.rotate(a)
                painter.drawText(QtCore.QPointF(bw2, -5.0), buf)
        else:
            painter.drawText(QtCore.QPointF(bw2, -5.0), '')
        painter.restore()


class MacroNode(object):
    '''The MacroNode menu is a stripped down version of the Node class with
    the capability of using a "Layout Window" instead of a NodeMenu.
    '''

    def __init__(self, graph, pos):
        self._graph = graph

        self._id = id(self)

        self._label = ''
        self._scrollArea_layoutWindow = None
        self._layoutWindow = None
        self.newLayoutWindow(config=0)

        self._collapsed = False
        # for load/paste
        self._destined_collapse = False

        # connected node groups
        self._src_cn = []
        self._sink_cn = []
        self._encap_nodes = []
        self._encap_nodepos = []

        self._src = PortEdge(self, graph, MacroAPI, role=0)
        self._src.setPos(pos)

        # self._sink = PortEdge(self, graph, nodeIF=self._src._nodeIF,
        # nodeIFscroll=self._src._nodeIF_scrollArea)
        self._sink = PortEdge(self, graph, MacroAPI, role=1)
        self._sink.setPos(QtCore.QPointF(0, 100) + pos)
        self._sink_pos = QtCore.QPointF(0, 100) + pos

        self._face = PortEdge(self, graph, MacroAPI, role=2)
        self._face.setPos(pos)
        self._face.inportList = []
        self._face.outportList = []

        # save expanded positions
        self._src_exppos = self._src.getPos()
        self._sink_exppos = self._sink.getPos()
        self._face_colpos = self._face.getPos()

        self._macroedge = MacroNodeEdge(self._src, self._sink)

        self._graph.scene().addItem(self._src)
        self._graph.scene().addItem(self._sink)
        self._graph.scene().addItem(self._macroedge)
        self._graph.scene().addItem(self._face)
        self._face.hide()

        self._anim_timeline = None
        self._anim = None

        self._scrollArea_layoutWindow.setWindowTitle('Macro')

    def getSettings(self):
        '''Keep all the settings required to instantiate the macro.
            -position, connectivity are all kept by normal serialization.
            -need collapse, label, layoutWindow.
        '''
        s = {}
        s['id'] = self.getID()
        s['label'] = self._label
        s['collapse'] = self.isCollapsed()
        s['layoutWindow'] = self._layoutWindow.getSettings()

        s['src_settings'] = self._src.getSettings()
        #s['src_pos'] = self._src_exppos
        s['sink_settings'] = self._sink.getSettings()
        #s['sink_pos'] = self._sink_exppos
        s['face_settings'] = self._face.getSettings()
        #s['face_pos'] = self._face_colpos

        x = self._sink_exppos[0] - self._src_exppos[0]
        y = self._sink_exppos[1] - self._src_exppos[1]
        s['rel_jaw_pos'] = [x, y]

        if self.isCollapsed():
            # save relative node positions in dict with id as the key
            # ids can be used at deserializing by lookup.
            npos = {}
            for node, rel in zip(self._encap_nodes, self._encap_nodepos):
                npos[str(node.getID())] = [rel.x(), rel.y()]
            s['nodes_rel_pos'] = npos

        return s

    def getNodeByID(self, buf, nid):
        for item in buf:
            if isinstance(item, Node):
                if item.getID() == nid:
                    return item

    def loadSettings(self, s, nodeList, pos):
        '''load all settings from a description generated by getSettings()
        '''
        self.setID(s['id'])
        self.setLabelWdg(s['label'])
        self.setLabel(s['label'])

        self.newLayoutWindowFromSettings(s['layoutWindow'], nodeList)

        if s['collapse']:
            x = s['face_settings']['pos'][0] + pos[0]
            y = s['face_settings']['pos'][1] + pos[1]
            #x = s['src_pos'][0] + pos[0]
            #y = s['src_pos'][1] + pos[1]
            self._src.setPos(QtCore.QPointF(x, y))

            # need relative jaw pos
            #x = s['sink_settings']['pos'][0] + pos[0]
            #y = s['sink_settings']['pos'][1] + pos[1]
            #self._sink.setPos(QtCore.QPoint(x, y))
            x += s['rel_jaw_pos'][0]
            y += s['rel_jaw_pos'][1]
            self._sink.setPos(QtCore.QPointF(x, y))

            x = s['face_settings']['pos'][0] + pos[0]
            y = s['face_settings']['pos'][1] + pos[1]
            self._face.setPos(QtCore.QPointF(x, y))

            rel = QtCore.QPointF(x, y)
            for nid, epos in list(s['nodes_rel_pos'].items()):
                enode = self.getNodeByID(nodeList, int(nid))
                if enode:
                    enode.setPos(rel + QtCore.QPointF(*epos))
                    log.debug("node found by nid: "+enode.name)
                    log.debug("pos: ")
                    log.debug(str(epos))
                    log.debug(str(rel))
                    log.debug(str(QtCore.QPointF(*epos)))
                    log.debug(str(rel + QtCore.QPointF(*epos)))
                else:
                    log.warn("nid not found: "+str(nid))

        else:
            x = s['src_settings']['pos'][0] + pos[0]
            y = s['src_settings']['pos'][1] + pos[1]
            self._src.setPos(QtCore.QPointF(x, y))

            x = s['sink_settings']['pos'][0] + pos[0]
            y = s['sink_settings']['pos'][1] + pos[1]
            self._sink.setPos(QtCore.QPointF(x, y))

            x = s['src_settings']['pos'][0] + pos[0]
            y = s['src_settings']['pos'][1] + pos[1]
            self._face.setPos(QtCore.QPointF(x, y))

        self._src.setID(s['src_settings']['id'])
        self._sink.setID(s['sink_settings']['id'])
        self._face.setID(s['face_settings']['id'])

        # save expanded positions
        self._src_exppos = self._src.getPos()
        self._sink_exppos = self._sink.getPos()
        self._face_colpos = self._face.getPos()

        # self.setCollapse(s['collapse'])
        self._destined_collapse = s['collapse']

    def setID(self, val=None):
        if val is None:
            self._id = id(self)
        else:
            self._id = val

    def getID(self):
        return self._id

    def resetPortParents(self):
        '''Return the parent status of each port to their respective _src or
        _sink origin nodes.
        '''
        for port in self._src.inportList:
            port.setParentItem(self._src)
            port.resetPos()
            port.show()

        for port in self._sink.outportList:
            port.setParentItem(self._sink)
            port.resetPos()
            port.show()

    def showFacePorts(self):
        '''Set the parent of each borrowed port to be the face node.
        '''
        self._face.inportList = []
        self._face.outportList = []

        for ip, op in zip(self._src.inportList, self._src.outportList):
            if ip.edgeCount() or op.edgeCount():
                ip.setParentItem(self._face)
                ip.setPosByPortNum(len(self._face.inportList))
                self._face.inportList.append(ip)
                ip.show()
            else:
                ip.hide()

        for op, ip in zip(self._sink.outportList, self._sink.inportList):
            if ip.edgeCount() or op.edgeCount():
                op.setParentItem(self._face)
                op.setPosByPortNum(len(self._face.outportList))
                self._face.outportList.append(op)
                op.show()
            else:
                op.hide()

        self._face.update()

    def toggleCollapse(self):
        if self.isCollapsed():
            self.setCollapse(False)
        else:
            self.setCollapse(True)

    def isCollapsed(self):
        return self._collapsed

    def setCollapse(self, val):
        self._collapsed = val
        if self.isCollapsed():
            # only set it if its truly collapsible.
            if not self.collapseMacro():
                self._collapsed = False
        else:
            self.expandMacro()

    def setCollapse_NoAction(self, val):
        '''Just set the flag
        '''
        self._collapsed = val

    def shouldCollapse(self):
        '''for load/save, should this be collapsed?
        '''
        return self._destined_collapse

    def getNodes(self):
        return [self._src, self._sink, self._face]

    def resetIDs(self):
        '''After the node has been loaded and reconnected to its neighbors,
        reset the ID for future copy/paste or load/save uniqueness.
        '''
        self.setID()
        self._src.setID()
        self._sink.setID()
        self._face.setID()

    def deleteMacroEdge(self):
        '''The Macro-edge connecting the Input and Output nodes.
        '''
        if self._macroedge.scene():
            self._graph.scene().removeItem(self._macroedge)

    def closeLayoutWidget(self):
        # self._layoutWindow.close()
        self._scrollArea_layoutWindow.close()

    def readyForDeletion(self):
        self.closeLayoutWidget()
        self.deleteMacroEdge()

        # delete all encapsulated nodes if its collapsed
        if self.isCollapsed():
            for node in self._encap_nodes:
                if node.scene():
                    self._graph.deleteNode(node)

    def newLayoutWindowFromSettings(self, s, nodeList):
        '''Copied from canvasGraph.
        '''

        layoutwindow = LayoutMaster(self._graph, config=s['config'], macro=True, labelWin=True)
        layoutwindow.loadSettings(s, nodeList)
        layoutwindow.setWindowTitle('Macro')
        layoutwindow.wdglabel.textChanged.connect(self.setLabel)

        self._scrollArea_layoutWindow = QtWidgets.QScrollArea()
        self._scrollArea_layoutWindow.setWidget(layoutwindow)
        self._scrollArea_layoutWindow.setWidgetResizable(True)
        self._scrollArea_layoutWindow.setGeometry(50, 50, 300, 1000)

        layoutwindow.setGeometry(50, 50, 400, 300)
        self._layoutWindow = layoutwindow

    def newLayoutWindow(self, config):
        '''Copied from canvasGraph.
        '''

        layoutwindow = LayoutMaster(self._graph, config=config, macro=True, labelWin=True)
        layoutwindow.setWindowTitle('Macro')
        layoutwindow.wdglabel.textChanged.connect(self.setLabel)

        self._scrollArea_layoutWindow = QtWidgets.QScrollArea()
        self._scrollArea_layoutWindow.setWidget(layoutwindow)
        self._scrollArea_layoutWindow.setWidgetResizable(True)
        self._scrollArea_layoutWindow.setGeometry(50, 50, 300, 1000)

        layoutwindow.setGeometry(50, 50, 400, 300)
        self._layoutWindow = layoutwindow

    def setLabel(self, lab):
        '''Set the internal label buffer and update the display names for each
        node and node-menu-window.
        '''
        self._label = str(lab)

        # update the node-menu window title
        if self._label != '':
            self._scrollArea_layoutWindow.setWindowTitle('Macro: ' + self._label)
        else:
            self._scrollArea_layoutWindow.setWindowTitle('Macro')

        self._src.update()
        self._sink.update()
        self._face.update()

    def setLabelWdg(self, lab):
        '''Sets the QLineEdit label directly for re-instantiating purposes.
        '''
        self._layoutWindow.wdglabel.setText(lab)

    def getEncapsulatedNodes(self):
        return self._encap_nodes

    def listEncapsulatedNodes(self):
        '''Searches for all nodes connected to the output of the _src and the
        input of the _sink.
        '''
        all_cn = self._sink_cn + self._src_cn

        enc_nodes = []
        for cn in all_cn:
            if cn[0] != self._src and cn[0] != self._sink:
                enc_nodes.append(cn[0])
            if cn[1] != self._src and cn[1] != self._sink:
                enc_nodes.append(cn[1])

        self._encap_nodes = list(set(enc_nodes))

    def isProcessing(self):
        '''Look at the run status of all nodes in the macro to determine if
        macro is running.
        '''
        for node in self._encap_nodes:
            if node.isProcessingEvent():
                return True
        if self._src.isProcessingEvent():
            return True
        if self._src.isProcessingEvent():
            return True
        return False

    def inComputeErrorState(self):
        '''Look at the run status of all nodes in the macro to determine if
        macro is in error.
        '''
        for node in self._encap_nodes:
            if node.inComputeErrorState():
                return True
        if self._src.inComputeErrorState():
            return True
        if self._src.inComputeErrorState():
            return True
        return False

    def inInitUIErrorState(self):
        '''Look at the run status of all nodes in the macro to determine if
        macro is in error.
        '''
        for node in self._encap_nodes:
            if node.inInitUIErrorState():
                return True
        if self._src.inInitUIErrorState():
            return True
        if self._src.inInitUIErrorState():
            return True
        return False

    def inValidateErrorState(self):
        '''Look at the run status of all nodes in the macro to determine if
        macro is in warning state.
        '''
        for node in self._encap_nodes:
            if node.inValidateErrorState():
                return True
        if self._src.inValidateErrorState():
            return True
        if self._src.inValidateErrorState():
            return True
        return False

    def validateSinkNodes(self):
        '''Get the list of connected nodes for _sink macro nodes.
        Determine if the connections are valid, then list the valid nodes.
        '''
        # get _sink connections
        # throw out input connections (where THIS node is the source).
        src_tmp = self._sink.getConnectionTuples()
        src_cn = [cn for cn in src_tmp if cn[0] != self._sink]
        # store output connections for comparison
        src_out = [cn for cn in src_tmp if cn[0] == self._sink]
        src_ill = src_out[:]

        log.debug("downstream search")
        while True:

            # build a local connection list for all nodes in the next level
            lc = []
            for cn in src_out:

                # sources
                if cn[0] != self._sink:  # don't backtrack
                    for outport in cn[0].outportList:
                        lc += outport.getConnectionTuples()

                # sinks
                for inport in cn[1].inportList:
                    lc += inport.getConnectionTuples()

                # cyclic
                if cn[1] == self._src:
                    log.debug("illegal cn (sink validate):")
                    log.debug("\t"+cn[0].name +"->"+cn[1].name)
                    return 1

            # consolidate connections
            pre_len = len(src_out)
            src_out += lc
            src_out = list(set(src_out))

            # if no new connections were added, then quit
            if len(src_out) == pre_len:
                break

        if manager.isDebug():
            log.debug("before upstream search")
            for cn in src_cn:
                log.debug("\t"+cn[0].name +"->"+cn[1].name)

        while True:

            # build a local connection list for all nodes in the next level
            lc = []
            for cn in src_cn:

                # sources
                if cn[0] != self._src:
                    for outport in cn[0].outportList:
                        lc += outport.getConnectionTuples()
                    for inport in cn[0].inportList:
                        lc += inport.getConnectionTuples()

                # sinks
                if cn[1] != self._sink:
                    for inport in cn[1].inportList:
                        lc += inport.getConnectionTuples()
                    for outport in cn[1].outportList:
                        lc += outport.getConnectionTuples()

                # cyclic
                if cn[1] == self._src:
                    log.debug("illegal cn (sink validate):")
                    log.debug("\t"+cn[0].name +"->"+cn[1].name)
                    return 1

            # consolidate connections
            pre_len = len(src_cn)
            src_cn += lc
            src_cn = list(set(src_cn))

            # if no new connections were added, then quit
            if len(src_cn) == pre_len:
                break

        if manager.isDebug():
            log.debug("after search:")
            for cn in src_cn:
                log.debug("\t"+cn[0].name +"->"+cn[1].name)

        # determine if the macro is valid
        log.debug("illegal macro connections:")
        ill_cnt = 0
        for cn in src_cn:
            if cn in src_ill:
                ill_cnt += 1
                log.debug("\t"+cn[0].name +"->"+cn[1].name)

        # check for any other macro nodes
        for cn in src_cn:
            if isMacroChildNode(cn[0]):
                if cn[0] not in self.getNodes():
                    ill_cnt += 1

            if isMacroChildNode(cn[1]):
                if cn[1] not in self.getNodes():
                    ill_cnt += 1

        self._sink_cn = src_cn

        return ill_cnt

    def validateSrcNodes(self):
        '''Get the list of connected nodes for _src macro nodes.
        Determine if the connections are valid, then list the valid nodes.
        '''

        # get _src connections
        # throw out input connections (where THIS node is the sink).
        src_tmp = self._src.getConnectionTuples()
        src_cn = [cn for cn in src_tmp if cn[1] != self._src]
        # store input connections for comparison
        src_in = [cn for cn in src_tmp if cn[1] == self._src]

        if manager.isDebug():
            log.debug("before search")
            for cn in src_cn:
                log.debug("\t"+cn[0].name +"->"+cn[1].name)

        while True:

            # build a local connection list for all nodes in the next level
            lc = []
            for cn in src_cn:

                # sources
                if len(cn[0].getConnectionTuples()):
                    if cn[0] != self._src:  # don't backtrack
                        lc += cn[0].getConnectionTuples()

                # cyclic
                if cn[0] == self._sink:
                    log.debug("illegal cn (src validate):")
                    log.debug("\t"+cn[0].name +"->"+cn[1].name)
                    return 1

                # sinks
                if cn[1] != self._sink:
                    if len(cn[1].getConnectionTuples()):
                        lc += cn[1].getConnectionTuples()

            # consolidate connections
            pre_len = len(src_cn)
            src_cn += lc
            src_cn = list(set(src_cn))

            # if no new connections were added, then quit
            if len(src_cn) == pre_len:
                break

        if manager.isDebug():
            log.debug("after search:")
            for cn in src_cn:
                log.debug("\t"+cn[0].name +"->"+cn[1].name)

        # determine if the macro is valid
        log.debug("illegal macro connections:")
        ill_cnt = 0
        for cn in src_cn:
            if cn in src_in:
                ill_cnt += 1
                log.debug("\t"+cn[0].name +"->"+cn[1].name)

        # check for any other macro nodes
        for cn in src_cn:
            if isMacroChildNode(cn[0]):
                if cn[0] not in self.getNodes():
                    ill_cnt += 1

            if isMacroChildNode(cn[1]):
                if cn[1] not in self.getNodes():
                    ill_cnt += 1

        self._src_cn = src_cn

        return ill_cnt

    def saveNodePos(self):
        '''Run thru the list of encapsulated nodes and save their original
        relative position to the _src node.
        '''
        self._sink_pos = self._sink.pos() - self._src.pos()

        # save node-pos with parallel list to _encap_nodes
        self._encap_nodepos = []
        for node in self._encap_nodes:
            self._encap_nodepos.append(node.pos() - self._src.pos())

        # save the expanded sibling node positions
        self._src_exppos = self._src.getPos()
        self._sink_exppos = self._sink.getPos()

        return self._encap_nodepos

    def collapseMacro(self):

        if self.validateSrcNodes():
            return False

        if self.validateSinkNodes():
            return False

        self.listEncapsulatedNodes()

        if manager.isDebug():
            log.debug("nodes:")
            for node in self._encap_nodes:
                log.debug(node.name)

        self.saveNodePos()

        self._face.setPos(self._src.pos())

        # initialize animation stuff (close jaw)
        self._anim = QtCore.QParallelAnimationGroup()
        self._anim.finished.connect(self.jawClosed)
        anim = QtCore.QPropertyAnimation(self._sink, b"pos")
        anim.setDuration(300)
        anim.setEndValue(self._src.pos())
        self._anim.addAnimation(anim)

        for node in self._encap_nodes:
            anim = QtCore.QPropertyAnimation(node, b"pos")
            anim.setDuration(300)
            anim.setEndValue(self._src.pos())
            self._anim.addAnimation(anim)

        self._anim.start()

        return True

    def jawClosed(self):
        '''finish the collapse Macro animation by hiding all encapsulated nodes
        -connect all nodes
        '''
        # hide all the nodes and connect their update-signals to the macro
        for node in self._encap_nodes:
            node.hide()
            node._forceUpdate.connect(self.updateNodePainter)
            for edge in node.edges():
                edge.hide()

        self._sink.hide()
        self._src.hide()
        self._macroedge.hide()
        self._face.show()

        self.showFacePorts()

        for port in self._face.getPorts():
            port.updateEdges()

        # make sure the face is selected
        self._face.setSelected(True)
        self._src.setSelected(False)
        self._sink.setSelected(False)

    def updateNodePainter(self):
        '''Call the node updates for each of the bound nodes for re-coloring.
        '''
        if self.isCollapsed():
            self._face.forceUpdate_NodeUI()

    def jawOpen(self):
        '''Once the open animation is finished, disconnect the re-painting
        updater signal.
        '''
        for node in self._encap_nodes:
            node._forceUpdate.disconnect()
            node.setSelected(True)

        # make sure the end nodes are also selected
        self._face.setSelected(False)
        self._src.setSelected(True)
        self._sink.setSelected(True)

    def expandMacro(self):

        # before expansion save the face position for load/paste purposes
        self._face_colpos = self._face.getPos()

        self._src.setPos(self._face.pos())
        self._sink.setPos(self._src.pos())

        self._face.hide()
        self._macroedge.show()
        self._sink.show()
        self._src.show()
        self.resetPortParents()

        for port in self._src.getPorts():
            port.updateEdges()

        for port in self._sink.getPorts():
            port.updateEdges()

        for node in self._encap_nodes:
            node.setPos(self._src.pos())
            node.show()
            for edge in node.edges():
                edge.show()

        # initialize animation (open jaw)
        self._anim = QtCore.QParallelAnimationGroup()
        self._anim.finished.connect(self.jawOpen)
        anim = QtCore.QPropertyAnimation(self._sink, b"pos")
        anim.setDuration(300)
        anim.setEndValue(self._sink_pos + self._src.pos())
        self._anim.addAnimation(anim)

        for node, pos in zip(self._encap_nodes, self._encap_nodepos):
            anim = QtCore.QPropertyAnimation(node, b"pos")
            anim.setDuration(300)
            anim.setEndValue(pos + self._src.pos())
            self._anim.addAnimation(anim)

        self._anim.start()



#!/usr/bin/env python

#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

PREFIX='/opt/anaconda1anaconda2anaconda3'

import os
import sys
import inspect

GPI_PKG=PREFIX
GPI_FRAMEWORK=GPI_PKG+'lib/'
sys.path.insert(0, GPI_FRAMEWORK) # gpi
sys.path.insert(0, GPI_PKG) # plugin

# for API related docs
from gpi import VERSION
from gpi import RELEASE_DATE
import gpi.nodeAPI
import gpi.widgets
import types
from types import *

# for node docs
from gpi.config import Config
from gpi.library import NodeCatalogItem
from gpi.catalog import Catalog
from gpi.defines import isGPIModFile


# Node docs
class NodeDocs(object):
    '''(Deprecated) For grabbing the Node documentation for of each node found
    in all connected libraries.  This is used for listing all the nodes
    available. '''

    def __init__(self):
        #self._docText = '\n\n'+34*'-'+' NODE DOCS '+35*'-'+'\n\n'
        self._docText = ''
        self._known_GPI_nodes = Catalog()

        for path in Config.GPI_LIBRARY_PATH:
            self.scanGPIModules(path, recursion_depth=2)

        self.extractDocs()

    def scanGPIModules(self, ipath, recursion_depth=1):
        ocnt = ipath.count('/')
        new_sys_paths = []
        for path, dn, fn in os.walk(ipath):
            # TODO: instead of checking for hidden svn dirs, just choose any hidden dir
            if (path.count('/') - ocnt <= recursion_depth) and not path.count('/.svn'):
                for fil in os.listdir(path):

                    fullpath = path+'/'+fil

                    if isGPIModFile(fullpath):

                        item = NodeCatalogItem(fullpath)
                        item.load()  # load check
                        if item.valid():
                            self._known_GPI_nodes.append(item)
                        else:
                            print(("Failed to load: "+str(item)))


    def __str__(self):
        #return str(self._known_GPI_nodes)
        return self._docText
         
    def extractDocs(self):
        # look for ExternalNode
        cur_lib = ''
        cur_sub = ''
        for item in sorted(list(self._known_GPI_nodes.values()), key=lambda x: x.key().lower()):
            if hasattr(item.mod, 'ExternalNode'):
                cur_doc = str(inspect.getdoc(getattr(item.mod, 'ExternalNode')))

                numSpaces = 4
                cur_doc = "\n".join((numSpaces * " ") + i for i in cur_doc.splitlines())

                #self._docText += '\n'+80*'*'+'\n' 
                #self._docText += '\n\\subsection{'+str(item.key()).replace('_', '\_') +'}' + '\n\n'


                if item.third != cur_lib:
                    self._docText += '\n# '+str(item.third)+'\n'
                    cur_lib = item.third

                if item.second != cur_sub:
                    self._docText += '\n## '+str(item.second)+'\n'
                    cur_sub = item.second


                self._docText += '\n### '+str(item.name).replace('_', '\_') +'\n\n'
                #self._docText += '\n\\begin{lstlisting}\n'
                self._docText += str(cur_doc)
                self._docText += '\n\n'
                #self._docText += '\n\\end{lstlisting}\n'
                #self._docText += '\n'+80*'*'+'\n' 

            else:
                print((str(item) + ' Doesnt have ExternalNode definition, skipping...'))




# API docs
class GPIdocs(object):
    '''(Deprecated) For gathering all the relevant API documentation used in
    Node development. '''

    def __init__(self):
        self._docText = None
        self.parmList = []

        self.generateHelpText()

    def __str__(self):
        #return str(self._known_GPI_nodes)
        return self._docText

    def generateHelpText(self):
        """Gather the __doc__ string of the ExternalNode derived class,
        all the set_ methods for each attached widget, and any attached
        GPI-types from each of the ports.
        """
        # WIDGETS DOC
        # parm_doc = ""  # contains parameter info
        wdg_doc = "\n\n# WIDGETS\n"  # generic widget ref info
        wdg_doc += '''
        \nThis is a list of widgets and their associated attributes to aid in
        parameterizing and declaring widget.
        '''
        for name in dir(gpi.widgets):
            obj = getattr(gpi.widgets, name)
            if hasattr(obj, 'GPIWdgType'):
                self.parmList.append([name, obj])

        # get the generic group to filter out base members
        generic_wdg = None
        for parm in self.parmList:
            if parm[0] == 'GenericWidgetGroup':
                generic_wdg = parm
                #print parm[1].__dict__

                wdg_doc += "\n###  " + parm[0] + "\n"
                numSpaces = 8
                set_doc = "\n".join((" ") + i for i in str(
                    parm[1].__doc__).splitlines())
                wdg_doc += set_doc + "\n"
                wdg_doc += "attribute | type | description\n"
                wdg_doc += "--- | --- | ---\n"
                # set methods
                for member in dir(parm[1]):
                    if member.startswith('set_'):
                        #wdg_doc += (8 * " ") + member + inspect.formatargspec(
                        #    *inspect.getargspec(getattr(parm[1], member)))
                        wdg_doc += (" ") + "" + member.split('set_')[1] + " | "

                        set_doc = str(inspect.getdoc(getattr(parm[1], member)))
                        numSpaces = 1
                        set_doc = " ".join((
                            numSpaces * " ") + i for i in set_doc.splitlines())
                        wdg_doc += "" + set_doc + "\n"

        # now do other members
        for parm in self.parmList:
            if parm[0] == 'GenericWidgetGroup':
                continue

            wdg_doc += "\n###  " + parm[0] + "\n"
            numSpaces = 8
            set_doc = "\n".join((" ") + i for i in str(
                parm[1].__doc__).splitlines())
            wdg_doc += set_doc + "\n"
            wdg_doc += "attribute | type | description\n"
            wdg_doc += "--- | --- | ---\n"
            # set methods
            for member in dir(parm[1]):
                if member.startswith('set_'):
                    if member not in generic_wdg[1].__dict__ or member == 'set_val':
                        #wdg_doc += (8 * " ") + member + inspect.formatargspec(
                        #    *inspect.getargspec(getattr(parm[1], member)))
                        wdg_doc += (" ") + "" + member.split('set_')[1] + " | "

                        set_doc = str(inspect.getdoc(getattr(parm[1], member)))
                        numSpaces = 1
                        set_doc = " ".join((
                            numSpaces * " ") + i for i in set_doc.splitlines())
                        wdg_doc += "" + set_doc + "\n"

        # PORTS DOC
        port_doc = "\n\n# PORT Types\n"  # get the port type info
        port_doc += '''
        \nThe following types are used to define a node port for limiting the
        connection between nodes to the predefined port-types.  These labels are
        used when implementing `addInPort()` and `addOutPort()` port declaration
        functions.

        \n### PASS
        \nHas the affinity for any port type and allows any port connection.
        '''

        gpitype_libs = []
        for name in dir(plugin):
            if name.endswith('_GPITYPE'):
                gpitype_libs.append(name)

        gpitypes = []
        for name in gpitype_libs:
            lib = getattr(plugin, name)
            for oname in dir(lib):
                obj = getattr(lib, oname)
                if hasattr(obj, 'GPIType') and (oname is not 'GPIDefaultType'):
                    gpitypes.append([oname, obj])

        #for port in self.node.getPorts():
        for port in gpitypes:
            #typ = port.GPIType()
            #port_doc += "\n  \'" + port.portTitle + "\': " + \
            #    str(typ.__class__) + "|" + str(type(port)) + "\n"

            port_doc += "\n###   " + port[0] + "\n"

            # description
            numSpaces = 8
            set_doc = "\n".join((" ") + i for i in str(
                port[1].__doc__).splitlines())
            port_doc += set_doc + "\n"

            set_cnt = 0
            for member in dir(port[1]):
                if member.startswith('set_'):
                    set_cnt += 1

            if set_cnt:
                port_doc += "attribute | type | description\n"
                port_doc += "--- | --- | ---\n"

            # set methods
            for member in dir(port[1]):
                if member.startswith('set_'):
                    #port_doc += (8 * " ") + member + inspect.formatargspec(
                    #    *inspect.getargspec(getattr(port[1], member)))
                    port_doc += (" ") + "" + member.split('set_')[1] + " | "

                    set_doc = str(inspect.getdoc(getattr(port[1], member)))
                    numSpaces = 16
                    set_doc = " ".join((" ") + i for i in set_doc.splitlines())
                    port_doc += "" + set_doc + "\n"

        port_doc += "\n"

        # GETTERS/SETTERS
        getset_doc = "\n\n# GETTERS & SETTERS\n\n"
        getset_doc += '''
        \nGetters & Setters are functions that make up the node API.  They provide
        access to the UI elements (i.e. in-ports, out-ports, widgets).
        '''
        getset_doc += "\n## Node Initialization\n\n"
        getset_doc += '''
        \nThese methods are used to declare in-ports, out-ports and widgets within
        the `initUI()` node method.
        '''
        getset_doc += self.formatFuncDoc(gpi.nodeAPI.NodeAPI.addWidget)
        getset_doc += self.formatFuncDoc(gpi.nodeAPI.NodeAPI.addInPort)
        getset_doc += self.formatFuncDoc(gpi.nodeAPI.NodeAPI.addOutPort)
        getset_doc += "\n## Node Compute\n\n"
        getset_doc += '''
        \nThese methods are used to reference data from in-ports, out-ports and widgets
        within the `compute()` routine.
        '''
        getset_doc += self.formatFuncDoc(gpi.nodeAPI.NodeAPI.getVal)
        getset_doc += self.formatFuncDoc(gpi.nodeAPI.NodeAPI.getAttr)
        getset_doc += self.formatFuncDoc(gpi.nodeAPI.NodeAPI.setAttr)
        getset_doc += self.formatFuncDoc(gpi.nodeAPI.NodeAPI.getData)
        getset_doc += self.formatFuncDoc(gpi.nodeAPI.NodeAPI.setData)

        #self._docText = node_doc  # + wdg_doc + port_doc + getset_doc
        self._docText = wdg_doc + port_doc + getset_doc

        #self.wdgabout.setPlainText(self._docText)


    def formatFuncDoc(self, func):
        """Generate auto-doc for passed func obj."""
        numSpaces = 24
        fdoc = inspect.getdoc(func)
        set_doc = "\n".join((
            numSpaces * " ") + i for i in str(fdoc).splitlines())
        rdoc = "\n### "+func.__name__+ "\n" + (16 * " ") + func.__name__ + \
            inspect.formatargspec(*inspect.getargspec(func)) \
            + "\n" + set_doc + "\n\n"
        return rdoc





if __name__ == '__main__':

    api_docs = GPIdocs()
    #print api_docs

    node_docs = NodeDocs()
    #print node_docs

    boiler = '<b>This file was auto-generated via the docs.py script (GPI '+VERSION+', '+RELEASE_DATE+').'+'  Do not edit this file directly.</b>\n\n'

    # NodeAPI
    with open('NodeAPI.md','w') as f:
        print("Writing to NodeAPI.md...")

        preamble = '''# Node API\n'''

        f.write(boiler)
        f.write(preamble)
        f.write(str(api_docs))


    # CoreNodes
    with open('CoreNodes.md','w') as f:
        print("Writing to CoreNodes.md...")

        preamble = '''This is the core node library.
The following node usage information was generated from comments written into
the node-code by the authors.

'''
        f.write(boiler)
        f.write(preamble)
        f.write(str(node_docs))



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

'''This is an initial attempt at logging GPI sessions.  Currently all logging
information is printed to stdout b/c there is a missing mechanism to
communicate logging statements back from forked processes. '''

import time
import inspect
import logging

GPINODE = logging.INFO + 1
DIALOG = logging.CRITICAL + 1

# A more thread safe replacement for the std::logging package.
# At the time this was written there were errors in threaded modules:
#  Exception AttributeError: AttributeError("'_DummyThread' object has no attribute '_Thread__block'",) in <module 'threading' from '/opt/local/Library/Frameworks/Python.framework/Versions/2.7/lib/python2.7/threading.pyc'> ignored  
# Just using the print command to stdout works for this application without the
# added errors. -This will need to be fixed if log files are implemented.
class PrintLogger(object):
    def __init__(self, name):
        self._name = name
        self._level = logging.WARNING

    def setLevel(self, lev):
        self._level = lev

    def debug(self, msg):
        if self._level <= logging.DEBUG:
            self.printb(msg, 'DEBUG')

    def info(self, msg):
        if self._level <= logging.INFO:
            self.printb(msg, 'INFO')

    def node(self, msg):
        if self._level <= GPINODE:
            self.printb(msg, 'NODE')

    def warn(self, msg):
        if self._level <= logging.WARNING:
            self.printb(msg, 'WARNING')

    def error(self, msg):
        if self._level <= logging.ERROR:
            self.printb(msg, 'ERROR')

    def critical(self, msg):
        if self._level <= logging.CRITICAL:
            self.printb(msg, 'CRITICAL')

    def dialog(self, msg):
        # things marked with dialog should eventually get a popup dialog box
        # to actually interact with.  This is just a reminder.
        if self._level <= DIALOG:
            self.printb(msg, '----')

    def printb(self, msg, lev):
        # get the line number by getting the current frame (printb()) then
        # going back to (debug(), info(), warn(), etc...), then back once
        # more to where the logger was called.
        lineno = str(inspect.currentframe().f_back.f_back.f_lineno)

        # the user input 'msg' is forced to be a string
        print((time.asctime(time.localtime()) + ' - ' + self._name + ':' + lineno + ' - ' + lev + ' - ' + str(msg)))


class GPILogManager(object):
    def __init__(self):
        # From: http://docs.python.org/2/howto/logging.html

        # default level, keep loggers on full
        self.default_level = logging.DEBUG

        # default level, start console on mid
        self.console_level = logging.WARNING

        # create logger
        self.logger = logging.getLogger(__name__)
        self.logger.setLevel(self.default_level)

        # create console handler and set level to debug
        #self.ch = logging.NullHandler()  # Turn completely off
        self.ch = logging.StreamHandler()
        self.ch.setLevel(self.console_level)

        # create formatter
        self.formatter = logging.Formatter(
            '%(asctime)s - %(name)s:%(lineno)d - %(levelname)s - %(message)s')

        # add formatter to ch
        self.ch.setFormatter(self.formatter)

        # add ch to logger
        self.logger.addHandler(self.ch)

    def getLogger(self, name):
        # automatically attach it to the console handle
        logger = logging.getLogger(name)
        logger.addHandler(self.ch)
        logger.setLevel(self.default_level)
        return logger

    def setLevel(self, lev):
        # set the console level
        self.ch.setLevel(lev)
        self.console_level = lev

    def isDebug(self):
        return self.console_level <= logging.DEBUG
    def isWarning(self):
        return self.console_level <= logging.WARNING
    def isInfo(self):
        return self.console_level <= logging.INFO
    def isError(self):
        return self.console_level <= logging.ERROR
    def isCritical(self):
        return self.console_level <= logging.CRITICAL

    def getMainLogger(self):
        return self.logger


class GPIPrintManager(GPILogManager):

    def __init__(self):
        super(GPIPrintManager, self).__init__()
    
        self.logger = PrintLogger(__name__)
        self.logger.setLevel(self.console_level)

        self.loggers = []

    def getLogger(self, name):
        logger = PrintLogger(name)
        logger.setLevel(self.console_level)
        self.loggers.append(logger)
        return logger

    def setLevel(self, lev):
        super(GPIPrintManager, self).setLevel(lev)

        for logger in self.loggers:
            logger.setLevel(lev)

    def getMainLogger(self):
        return self.logger

# global handles
#manager = GPILogManager()
manager = GPIPrintManager()
logger = manager.getMainLogger()

# default log level
manager.setLevel(logging.WARNING)

# 'application' code
# logger.debug('debug message')
# logger.info('info message')
# logger.warn('warn message')
# logger.error('error message')
# logger.critical('critical message')
#
# manager.setLevel(logging.CRITICAL)
#
# logger.info('after switch')
#
# manager.setLevel(logging.DEBUG)



#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

import multiprocessing
from pathos.multiprocessing import Pool

class Parallel():
    def __init__(self):
        self.pool = None
        self.processes = multiprocessing.cpu_count() 

    def set_processes(self, processes):
        self.processes = processes

    def start(self):
        self.pool = Pool(processes=self.processes)
    
    def function(self, function):
        def wrapper(*args):
            if self.pool is None: self.start()
            promise = self.pool.apply_async(function, args)
            return promise
        return wrapper

    def get(self, promises):
        return list(map(lambda promise: promise.get(), promises))

    def close(self):
        if self.pool is None: return

        # terminate proccesses
        self.pool.close()
        self.pool.join()



#    Copyright (C) 2014  Dignity Health
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
#    NO CLINICAL USE.  THE SOFTWARE IS NOT INTENDED FOR COMMERCIAL PURPOSES
#    AND SHOULD BE USED ONLY FOR NON-COMMERCIAL RESEARCH PURPOSES.  THE
#    SOFTWARE MAY NOT IN ANY EVENT BE USED FOR ANY CLINICAL OR DIAGNOSTIC
#    PURPOSES.  YOU ACKNOWLEDGE AND AGREE THAT THE SOFTWARE IS NOT INTENDED FOR
#    USE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITY, INCLUDING BUT NOT
#    LIMITED TO LIFE SUPPORT OR EMERGENCY MEDICAL OPERATIONS OR USES.  LICENSOR
#    MAKES NO WARRANTY AND HAS NO LIABILITY ARISING FROM ANY USE OF THE
#    SOFTWARE IN ANY HIGH RISK OR STRICT LIABILITY ACTIVITIES.

from .logger import manager
# start logger for this module
log = manager.getLogger(__name__)

import qtpy
import qtpy.QtCore as _QtCore
API_NAME = qtpy.API_NAME
API_NAME = API_NAME.split(' ')[0]  # truncate e.g. 'PyQt4 (API v2)' to 'PyQt4'

QtCore = _QtCore

def import_module(moduleName):
    p = __import__('qtpy', globals(), locals(), [moduleName], 0)
    return getattr(p, moduleName)

# QtWebKit or QtWebEngine may not be available. The modules will be set to None
# in this case. QtOpenGL is also deprecated in modern Qt.
def import_optional_module(moduleName):
    try:
        p = __import__(API_NAME, globals(), locals(), [moduleName], 0)
        return getattr(p, moduleName)
    except (AttributeError,ImportError) as e:
        return None

QtOpenGL = import_optional_module('QtOpenGL')
QtWebKit = import_optional_module('QtWebKit')
QtWebKitWidgets = import_optional_module('QtWebKitWidgets')
QtWebEngineWidgets = import_optional_module('QtWebEngineWidgets')

try:
    # PyQt4 and PySide
    QWebView = QtWebKit.QWebView
except AttributeError:
    QWebView = None
    if QtWebKitWidgets is not None:
        try:
            # PyQt5 and PySide2
            QWebView = QtWebKitWidgets.QWebView
        except:
            pass

Signal = QtCore.Signal
Slot = QtCore.Slot
Property = QtCore.Property
